# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# pylint: disable=missing-function-docstring

import torch
from lhotse.cut import Cut, MixedCut

from nemo.collections.common.data.lhotse.text_adapters import NeMoSFTExample, SourceTargetTextExample
from nemo.collections.common.data.prompt_fn import registered_prompt_format_fn
from nemo.collections.common.prompts.formatter import BOS_SLOT, EOS_SLOT, Modality, PromptFormatter


class Llama2PromptFormatter(PromptFormatter):
    """
    This template has been validated to provide identical tokenized results to the official code
    in https://github.com/meta-llama/llama/blob/main/llama/generation.py
    """

    NAME = "llama2"
    OUTPUT_ROLE = "assistant"
    TEMPLATE = {
        "system_and_user": {
            "template": f"{BOS_SLOT}[INST] <<SYS>>\n|system|\n<</SYS>>\n\n|message| [/INST]",
            "slots": {
                "system": Modality.Text,
                "message": Modality.Text,
            },
        },
        "user": {
            "template": f"{BOS_SLOT}[INST] |message| [/INST]",
            "slots": {
                "message": Modality.Text,
            },
        },
        OUTPUT_ROLE: {
            "template": f"|message| {EOS_SLOT}",
            "slots": {
                "message": Modality.Text,
            },
        },
    }


@registered_prompt_format_fn(Cut, Llama2PromptFormatter)
def llama2(cut: Cut, prompt: Llama2PromptFormatter) -> dict[str, torch.Tensor]:
    if isinstance(cut, MixedCut):
        cut = cut.first_non_padding_cut
    if cut.has_custom("context"):
        context = cut.context
    elif cut.has_custom("question"):
        context = cut.question
    else:
        context = cut.default_context

    turns = []
    if cut.has_custom("system_prompt"):
        turns.append({"role": "system_and_user", "slots": {"system": cut.system_prompt, "message": context}})
    else:
        turns.append({"role": "user", "slots": {"message": context}})
    if (answer := cut.supervisions[0].text) is not None:
        turns.append({"role": "assistant", "slots": {"message": answer}})
    return prompt.encode_dialog(turns)


@registered_prompt_format_fn(SourceTargetTextExample, Llama2PromptFormatter)
def llama2_src_tgt_text_example(example: SourceTargetTextExample, prompt: Llama2PromptFormatter):
    if example.question is not None:
        user_turn = {
            "role": "system_and_user",
            "slots": {"system": example.question.text, "message": example.source.text},
        }
    else:
        user_turn = {
            "role": "user",
            "slots": {"message": example.source.text},
        }
    return prompt.encode_dialog(
        [
            user_turn,
            {"role": prompt.OUTPUT_ROLE, "slots": {"message": example.target.text}},
        ]
    )


@registered_prompt_format_fn(NeMoSFTExample, Llama2PromptFormatter)
def llama2_sft_text_example(example: NeMoSFTExample, prompt: Llama2PromptFormatter):
    first_user_turn = example.data["conversations"][0]["value"]
    if "system" in example.data and example.data["system"]:
        first_turn = {
            "role": "system_and_user",
            "slots": {"system": example.data["system"], "message": first_user_turn},
        }
    else:
        first_turn = {
            "role": "user",
            "slots": {"message": first_user_turn},
        }
    return prompt.encode_dialog(
        [first_turn]
        + [
            {"role": "user" if turn["from"] == "User" else prompt.OUTPUT_ROLE, "slots": {"message": turn["value"]}}
            for turn in example.data["conversations"][1:]
        ]
    )


LLAMA3_BOS = "<|begin_of_text|>"
LLAMA3_HEADER_BEGIN = "<|start_header_id|>"
LLAMA3_HEADER_END = "<|end_header_id|>"
LLAMA3_END_OF_TURN = "<|eot_id|>"
LLAMA3_NL = "\n\n"


class Llama3PromptFormatter(PromptFormatter):
    """
    Implemented following the code at:
     https://github.com/meta-llama/llama3/blob/main/llama/test_tokenizer.py#L56
    """

    NAME = "llama3"
    OUTPUT_ROLE = "assistant"
    INFERENCE_PREFIX = f"{LLAMA3_HEADER_BEGIN}assistant{LLAMA3_HEADER_END}{LLAMA3_NL}"
    TEMPLATE = {
        "preamble": {
            "template": LLAMA3_BOS,
        },
        "system": {
            "template": f"{LLAMA3_HEADER_BEGIN}system{LLAMA3_HEADER_END}{LLAMA3_NL}|message|{LLAMA3_END_OF_TURN}",
            "slots": {
                "message": Modality.Text,
            },
        },
        "user": {
            "template": f"{LLAMA3_HEADER_BEGIN}user{LLAMA3_HEADER_END}{LLAMA3_NL}|message|{LLAMA3_END_OF_TURN}",
            "slots": {
                "message": Modality.Text,
            },
        },
        OUTPUT_ROLE: {
            "template": f"{INFERENCE_PREFIX}|message|{LLAMA3_END_OF_TURN}",
            "slots": {
                "message": Modality.Text,
            },
        },
    }


@registered_prompt_format_fn(Cut, Llama3PromptFormatter)
def llama3(cut: Cut, prompt: Llama3PromptFormatter) -> dict[str, torch.Tensor]:
    if isinstance(cut, MixedCut):
        cut = cut.first_non_padding_cut
    if cut.has_custom("context"):
        context = cut.context
    elif cut.has_custom("question"):
        context = cut.question
    else:
        context = cut.default_context

    turns = []
    if cut.has_custom("system_prompt"):
        turns.append({"role": "system", "slots": {"message": cut.system_prompt}})
    turns.append({"role": "user", "slots": {"message": context}})
    if (answer := cut.supervisions[0].text) is not None:
        turns.append({"role": "assistant", "slots": {"message": answer}})
    return prompt.encode_dialog(turns)



# Copyright (c) 2020, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import tempfile
from unittest import TestCase

import pytest
import torch
from omegaconf import DictConfig

from nemo.collections.asr.models import EncDecSpeakerLabelModel


class EncDecSpeechLabelModelTest(TestCase):
    @pytest.mark.unit
    def test_constructor(self):
        preprocessor = {
            '_target_': 'nemo.collections.asr.modules.AudioToMelSpectrogramPreprocessor',
        }
        encoder = {
            '_target_': 'nemo.collections.asr.modules.ConvASREncoder',
            'feat_in': 64,
            'activation': 'relu',
            'conv_mask': True,
            'jasper': [
                {
                    'filters': 512,
                    'repeat': 1,
                    'kernel': [1],
                    'stride': [1],
                    'dilation': [1],
                    'dropout': 0.0,
                    'residual': False,
                    'separable': False,
                }
            ],
        }

        decoder = {
            '_target_': 'nemo.collections.asr.modules.SpeakerDecoder',
            'feat_in': 512,
            'num_classes': 2,
            'pool_mode': 'xvector',
            'emb_sizes': [1024],
        }

        modelConfig = DictConfig(
            {
                'preprocessor': DictConfig(preprocessor),
                'encoder': DictConfig(encoder),
                'decoder': DictConfig(decoder),
            },
        )
        speaker_model = EncDecSpeakerLabelModel(cfg=modelConfig)
        speaker_model.train()
        # TODO: make proper config and assert correct number of weights

        # Check to/from config_dict:
        confdict = speaker_model.to_config_dict()
        instance2 = EncDecSpeakerLabelModel.from_config_dict(confdict)
        self.assertTrue(isinstance(instance2, EncDecSpeakerLabelModel))

    @pytest.mark.unit
    def test_ecapa_enc_dec(self):
        preprocessor = {
            '_target_': 'nemo.collections.asr.modules.AudioToMelSpectrogramPreprocessor',
        }
        encoder = {
            '_target_': 'nemo.collections.asr.modules.ECAPAEncoder',
            'feat_in': 80,
            'filters': [4, 4, 4, 4, 3],
            'kernel_sizes': [5, 3, 3, 3, 1],
            'dilations': [1, 1, 1, 1, 1],
            'scale': 2,
        }

        decoder = {
            '_target_': 'nemo.collections.asr.modules.SpeakerDecoder',
            'feat_in': 3,
            'num_classes': 2,
            'pool_mode': 'attention',
            'emb_sizes': 192,
        }

        modelConfig = DictConfig(
            {
                'preprocessor': DictConfig(preprocessor),
                'encoder': DictConfig(encoder),
                'decoder': DictConfig(decoder),
            }
        )
        speaker_model = EncDecSpeakerLabelModel(cfg=modelConfig)
        speaker_model.train()
        # TODO: make proper config and assert correct number of weights

        # Check to/from config_dict:
        confdict = speaker_model.to_config_dict()
        instance2 = EncDecSpeakerLabelModel.from_config_dict(confdict)
        self.assertTrue(isinstance(instance2, EncDecSpeakerLabelModel))

    @pytest.mark.unit
    def test_titanet_enc_dec(self):
        preprocessor = {
            '_target_': 'nemo.collections.asr.modules.AudioToMelSpectrogramPreprocessor',
        }
        encoder = {
            '_target_': 'nemo.collections.asr.modules.ConvASREncoder',
            'feat_in': 64,
            'activation': 'relu',
            'conv_mask': True,
            'jasper': [
                {
                    'filters': 256,
                    'repeat': 1,
                    'kernel': [1],
                    'stride': [1],
                    'dilation': [1],
                    'dropout': 0.0,
                    'residual': False,
                    'separable': True,
                    'se': True,
                    'se_context_size': -1,
                }
            ],
        }

        decoder = {
            '_target_': 'nemo.collections.asr.modules.SpeakerDecoder',
            'feat_in': 256,
            'num_classes': 2,
            'pool_mode': 'attention',
            'emb_sizes': [1024],
        }

        modelConfig = DictConfig(
            {
                'preprocessor': DictConfig(preprocessor),
                'encoder': DictConfig(encoder),
                'decoder': DictConfig(decoder),
            }
        )
        speaker_model = EncDecSpeakerLabelModel(cfg=modelConfig)
        speaker_model.train()
        # TODO: make proper config and assert correct number of weights

        # Check to/from config_dict:
        confdict = speaker_model.to_config_dict()
        instance2 = EncDecSpeakerLabelModel.from_config_dict(confdict)
        self.assertTrue(isinstance(instance2, EncDecSpeakerLabelModel))


class TestEncDecSpeechLabelModel:
    @pytest.mark.unit
    def test_pretrained_titanet_embeddings(self, test_data_dir):
        model_name = 'titanet_large'
        speaker_model = EncDecSpeakerLabelModel.from_pretrained(model_name)
        assert isinstance(speaker_model, EncDecSpeakerLabelModel)
        relative_filepath = "an4_speaker/an4/wav/an4_clstk/fash/an251-fash-b.wav"
        filename = os.path.join(test_data_dir, relative_filepath)

        emb, logits = speaker_model.infer_file(filename)

        class_id = logits.argmax(axis=-1)
        emb_sum = emb.sum()

        assert 11144 == class_id
        assert (emb_sum + 0.2575) <= 1e-2

    @pytest.mark.unit
    def test_pretrained_ambernet_logits(self, test_data_dir):
        model_name = 'langid_ambernet'
        lang_model = EncDecSpeakerLabelModel.from_pretrained(model_name)
        assert isinstance(lang_model, EncDecSpeakerLabelModel)
        relative_filepath = "an4_speaker/an4/wav/an4_clstk/fash/an255-fash-b.wav"
        filename = os.path.join(test_data_dir, relative_filepath)

        label = lang_model.get_label(filename)

        assert label == "en"

    @pytest.mark.unit
    def test_pretrained_ambernet_logits_batched(self, test_data_dir):
        model_name = 'langid_ambernet'
        lang_model = EncDecSpeakerLabelModel.from_pretrained(model_name)
        relative_filepath = "an4_speaker/an4/wav/an4_clstk/fash/an255-fash-b.wav"
        filename = os.path.join(test_data_dir, relative_filepath)
        device = 'cuda' if torch.cuda.is_available() else 'cpu'

        with tempfile.TemporaryDirectory() as tmpdir:
            temp_manifest = os.path.join(tmpdir, 'manifest.json')
            with open(temp_manifest, 'w', encoding='utf-8') as fp:
                entry = {"audio_filepath": filename, "duration": 4.5, "label": 'en'}
                fp.write(json.dumps(entry) + '\n')
                entry = {
                    "audio_filepath": filename,
                    "duration": 4.5,
                    "label": 'test',
                }  # test sample outside of training set
                fp.write(json.dumps(entry) + '\n')

            embs, logits, gt_labels, trained_labels = lang_model.batch_inference(temp_manifest, device=device)
            pred_label = trained_labels[logits.argmax(axis=-1)[0]]
            true_label = gt_labels[0]

            assert pred_label == true_label
            assert gt_labels[1] == 'test'



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from functools import lru_cache
from typing import Callable

import numpy as np

from nemo.collections.asr.inference.streaming.state.state import StreamingState
from nemo.collections.asr.inference.utils.constants import (
    POST_WORD_PUNCTUATION,
    ROUND_PRECISION,
    SENTENCEPIECE_UNDERSCORE,
)
from nemo.collections.asr.inference.utils.text_segment import TextSegment, Word
from nemo.collections.common.tokenizers.tokenizer_spec import TokenizerSpec


class BPEDecoder:
    """
    BPEDecoder class for decoding BPE (Byte Pair Encoding) tokens into words and segments by preserving timestamps and confidence scores
    """

    def __init__(
        self,
        vocabulary: list[str],
        tokenizer: TokenizerSpec,
        confidence_aggregator: Callable,
        asr_supported_puncts: set,
        word_boundary_tolerance: float,
        token_duration_in_secs: float,
    ):
        """
        Initialize the BPEDecoder.
        Args:
            vocabulary (list[str]): List of vocabulary tokens.
            tokenizer (TokenizerSpec): Tokenizer object.
            confidence_aggregator (Callable): Confidence aggregator function.
            asr_supported_puncts (set): Set of supported punctuation symbols.
            word_boundary_tolerance (float): Word boundary tolerance for timestamp refinement.
            token_duration_in_secs (float): Token duration in seconds.
        """

        self.vocabulary = vocabulary
        self.tokenizer = tokenizer
        self.confidence_aggregator = confidence_aggregator
        self.asr_supported_puncts = asr_supported_puncts
        self.punct_marks_with_underscore = asr_supported_puncts.union({SENTENCEPIECE_UNDERSCORE})
        self.word_boundary_tolerance = word_boundary_tolerance
        self.token_duration_in_secs = token_duration_in_secs
        self.start_of_word_cache = {
            token_id: token.startswith(SENTENCEPIECE_UNDERSCORE) for token_id, token in enumerate(self.vocabulary)
        }
        self.punct_cache = {
            token_id: (token in self.asr_supported_puncts, token in self.punct_marks_with_underscore)
            for token_id, token in enumerate(self.vocabulary)
        }

    @lru_cache(maxsize=10000)
    def cached_ids_to_text(self, tokens_slice: tuple[int]) -> str:
        """
        Cached tokenizer output to avoid repeated calls to the tokenizer.
        Args:
            tokens_slice (tuple): Tuple of token indices to be detokenized.
        Returns:
            str: Detokenized text.
        """
        word_text = self.tokenizer.ids_to_text(list(tokens_slice)).strip()
        return word_text

    def decode_bpe_tokens(self, state: StreamingState) -> None:
        """
        Decodes BPE tokens into words or segments with timestamps and confidence scores.
        Args:
            state (StreamingState): The state object containing the BPE tokens, timestamps, and confidence scores.
        """
        if state.options.is_word_level_output():
            # Form words and push them to the state
            decoded_words, need_merge = self.group_tokens_into_words(state.tokens, state.timesteps, state.confidences)
            state.push_back_words(decoded_words, need_merge, self.confidence_aggregator)
        elif state.options.is_segment_level_output():
            # Form text segment and push it to the state
            if state.tokens:
                decoded_segment, need_merge = self.group_tokens_into_segment(
                    state.tokens, state.timesteps, state.confidences
                )
                state.push_back_segment(decoded_segment, need_merge, self.confidence_aggregator)
        else:
            raise ValueError(f"Invalid output granularity: {state.options.asr_output_granularity}")

    def group_tokens_into_segment(
        self, tokens: list, timesteps: list, confidences: list
    ) -> tuple[TextSegment | None, bool]:
        """
        Group tokens into a text segment with timestamps and confidence scores.
        Args:
            tokens (list): List of token indices.
            timesteps (list): List of token timestamps.
            confidences (list): List of token confidence scores.
        Returns:
            (tuple[TextSegment | None, bool]) Text segment with text, start time, end time, and confidence score.
            Also returns a boolean to indicate if the text segment should be merged with the last segment stored in the state
        """
        n_tokens = len(tokens)

        if n_tokens != len(timesteps) or n_tokens != len(confidences):
            raise ValueError("tokens, timesteps and confidences must have the same length")

        if n_tokens == 0:
            return None, False

        need_merge = not bool(self.start_of_word_cache[tokens[0]])

        # Get the segment text
        segment_text = self.tokenizer.ids_to_text(tokens).strip()

        # Refine the start and end timestamps of the text segment
        start, end = self.refine_text_segment_timestamp(tokens, timesteps)

        # Convert token timestamps to seconds
        start = round(start * self.token_duration_in_secs, ROUND_PRECISION)
        end = round(end * self.token_duration_in_secs, ROUND_PRECISION)

        # Aggregate the confidence score of the text segment
        conf = self.confidence_aggregator(confidences)

        # Create a text segment
        return TextSegment(text=segment_text, start=start, end=end, conf=conf), need_merge

    def group_tokens_into_words(self, tokens: list, timesteps: list, confidences: list) -> tuple[list[Word], bool]:
        """
        Decodes BPE tokens into words with timestamps and confidence scores.
        Args:
            tokens (list): List of token indices.
            timesteps (list): List of token timesteps.
            confidences (list): List of token confidence scores.
        Returns:
            (tuple[list[Word], bool]) List of decoded words with text, start time, end time, and confidence score.
            Also returns a boolean to indicate if the first word should be merged with the last word stored in the state
        """
        n_tokens = len(tokens)

        if n_tokens != len(timesteps) or n_tokens != len(confidences):
            raise ValueError("tokens, timesteps and confidences must have the same length")

        if n_tokens == 0:
            return [], False

        # Group tokens into words
        is_start_mask = np.fromiter((self.start_of_word_cache[tok] for tok in tokens), dtype=np.int32)
        word_ids = np.cumsum(is_start_mask)

        start_indices = np.nonzero(np.diff(word_ids, prepend=word_ids[0] - 1))[0]
        end_indices = np.append(start_indices[1:], n_tokens)

        decoded_words, prev_word_end = [], None

        # If the first word is the start of a word, we need to merge it with the last word stored in the state
        need_merge = not bool(is_start_mask[0])

        for start_idx, end_idx in zip(start_indices, end_indices):

            tokens_slice = tokens[start_idx:end_idx]
            time_slice = timesteps[start_idx:end_idx]
            conf_slice = confidences[start_idx:end_idx]

            word_text = self.cached_ids_to_text(tuple(tokens_slice))

            # Ignore empty text
            if not word_text:
                continue

            # Append the post word punctuation to the previous word
            if word_text in POST_WORD_PUNCTUATION and len(decoded_words) > 0:
                prev_word = decoded_words[-1]
                prev_word.text += word_text
                continue

            # Refine timestamps
            word_start_tms, word_end_tms = self.refine_text_segment_timestamp(
                current_tokens=tokens_slice,
                current_timesteps=time_slice,
                next_segment_start_timestep=timesteps[end_idx] if end_idx < n_tokens else None,
                need_merge_with_next_segment=(
                    self.start_of_word_cache[tokens[end_idx]] if end_idx < n_tokens else None
                ),
                prev_segment_end=prev_word_end,
            )
            prev_word_end = word_end_tms

            # Aggregate confidence
            word_conf = self.confidence_aggregator(conf_slice)

            # Convert token timestamps to seconds
            start_sec = round(word_start_tms * self.token_duration_in_secs, ROUND_PRECISION)
            end_sec = round(word_end_tms * self.token_duration_in_secs, ROUND_PRECISION)

            decoded_words.append(Word(text=word_text, start=start_sec, end=end_sec, conf=word_conf))

        return decoded_words, need_merge

    def refine_text_segment_timestamp(
        self,
        current_tokens: list[int],
        current_timesteps: list[float],
        next_segment_start_timestep: float | None = None,
        need_merge_with_next_segment: bool | None = None,
        prev_segment_end: float | None = None,
    ) -> tuple[float, float]:
        """
        Refines the text segment timestamp based on the current tokens, timestamps, and the next segment start timestamp.
        Args:
            current_tokens (list[int]): List of token indices.
            current_timesteps (list[float]): List of token timestamps.
            next_segment_start_timestep (float | None): The start timestamp of the next segment.
            need_merge_with_next_segment (bool | None): True if the current segment should be merged with the next segment.
            prev_segment_end (float | None): The end timestamp of the previous segment.
        Returns:
            tuple(float, float): The refined start and end timestamps.
        """

        start, end = current_timesteps[0], current_timesteps[-1]

        # --- Correct the start timestamp if the first token is underscore or punctuation ---
        first_token = current_tokens[0]
        if self.punct_cache[first_token][1]:
            start = next(
                (tms for tms, token in zip(current_timesteps, current_tokens) if not self.punct_cache[token][1]),
                start,
            )

        # --- Correct the end timestamp if the last token is punctuation ---
        last_token = current_tokens[-1]
        if self.punct_cache[last_token][0]:
            end = next(
                (
                    current_timesteps[i]
                    for i in reversed(range(len(current_tokens)))
                    if not self.punct_cache[current_tokens[i]][0]
                ),
                end,
            )

        # --- If the next segment is close to the end of the current segment, merge timestamps ---
        if next_segment_start_timestep is not None and need_merge_with_next_segment:
            if next_segment_start_timestep - end <= self.word_boundary_tolerance:
                end = next_segment_start_timestep

        # --- Adjust the start and end timestamps based on the previous segment end ---
        delta = 0
        if prev_segment_end is not None:
            if prev_segment_end > start:
                delta = prev_segment_end - start

        start = start + delta
        end = end + delta
        return start, end + (1 if start == end else 0)



# Copyright (c) 2025, NVIDIA CORPORATION & AFFILIATES.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import tempfile
from pathlib import Path

import lightning.pytorch as pl
import numpy as np
import pytest
import torch
from lightning.pytorch.plugins.environments import TorchElasticEnvironment
from megatron.core import parallel_state
from omegaconf import DictConfig, OmegaConf

from nemo.collections.common.parts.nlp_overrides import NLPDDPStrategy
from nemo.collections.multimodal.speech_llm.models import modular_models
from nemo.collections.multimodal.speech_llm.parts.utils.data_utils import shift_tokens_by_multi_audios


class ModularAudioGPTModel(modular_models.ModularAudioGPTModel):
    # disable logging to avoid MisconfigurationException
    def log(self, *args, **kwargs):
        pass


def setup_module():
    pl.seed_everything(1)
    # init model parallel needed for LLM loss
    init_method = 'tcp://'
    master_ip = 'localhost'
    master_port = '6000'
    init_method += master_ip + ':' + master_port
    torch.distributed.init_process_group(backend='gloo', world_size=1, rank=0, init_method=init_method)
    parallel_state.initialize_model_parallel(1, 1)


@pytest.fixture
def llm_model_config():
    this_test_dir = os.path.dirname(os.path.abspath(__file__))
    # Although most of the stuff in model is loaded from ckpt, we need configs
    # for e.g. cfg.model.optim
    config = OmegaConf.load(
        os.path.join(
            this_test_dir,
            "../../../examples/multimodal/speech_llm/conf/modular_audio_gpt_config_peft.yaml",
        )
    )
    # TODO(zhehuai): move the following to Test /home/TestData
    config.model.restore_from_path = "/root/home/works/TestData/pretrained_models/megatron_gpt/gpt_pretrain_220m_len_4096_pos_alibi_step_595508_gbs256.nemo"
    config.model.micro_batch_size = 2
    config.model.global_batch_size = 2
    config.model.data.validation_ds.manifest_filepath = (
        '/root/home/works/TestData/datasets/LibriSpeech/dev_clean_cleaned.json'
    )
    config.model.data.train_ds.manifest_filepath = (
        '/root/home/works/TestData/datasets/LibriSpeech/dev_clean_cleaned.json'
    )
    return config


@pytest.fixture
def trainer_config():
    config_trainer = DictConfig({})

    if torch.cuda.is_available():
        accelerator = "gpu"
        torch.set_default_device('cuda')
    else:
        accelerator = "cpu"
    config_trainer.accelerator = accelerator
    config_trainer.devices = 1
    config_trainer.num_nodes = 1
    config_trainer.max_epochs = 4
    config_trainer.max_steps = 1
    config_trainer.val_check_interval = 1.0

    # for PyTorch Native AMP set precision=16
    config_trainer.precision = 32

    # setup cluster environment parameters"
    # use torch elastic cluster environment so `create_process_externally` is True
    # the launcher is set to None. It will not try to spawn new processes.
    # It won't create the misconfiguration error because of the `interactive session`
    os.environ["LOCAL_RANK"] = "0"
    os.environ["RANK"] = "0"
    os.environ["WORLD_SIZE"] = "1"

    strategy = NLPDDPStrategy()
    plugins = [TorchElasticEnvironment()]
    trainer = pl.Trainer(logger=False, plugins=plugins, strategy=strategy, **config_trainer)
    return trainer, config_trainer


@pytest.fixture
def perception_model_config():
    preprocessor = {"_target_": "nemo.collections.asr.modules.AudioToMelSpectrogramPreprocessor"}
    encoder = {
        "_target_": "nemo.collections.asr.modules.ConformerEncoder",
        "feat_in": 64,
        "n_layers": 8,
        "d_model": 64,
        "self_attention_model": "rel_pos_local_attn",
        "att_context_size": [128, 128],
    }

    model_config = DictConfig(
        {
            "_target_": "nemo.collections.multimodal.speechllm.modules.speechllm_perception.AudioPerceptionModule",
            "preprocessor": DictConfig(preprocessor),
            "encoder": DictConfig(encoder),
            "modality_adapter": DictConfig(encoder),
            "output_dim": 1024,
        }
    )
    return model_config


@pytest.fixture
def test_batch():
    signal_len = torch.from_numpy(np.array([64000, 64000]))
    transcript = torch.arange(10).reshape(2, 5).int()
    tokens = transcript[:, :-1]
    labels = transcript[:, 1:]
    transcript_length = torch.Tensor([3, 2]).int()
    # assuming context_lengths = [1, 1]
    loss_mask = torch.Tensor([[0, 1, 1, 0], [0, 1, 0, 0]])
    batch = {
        'audio_signal_length': signal_len,
        'tokens': tokens,
        'tokens_length': transcript_length,
        'contexts': torch.arange(260).reshape(2, 130).int(),
        'context_lengths': torch.Tensor([1, 1]).int(),
        'labels': labels,
        'answers': labels,
        'loss_mask': loss_mask,
    }
    batch['audio_signal'] = torch.randn([2, 64000])
    return batch


@pytest.mark.skip(reason="nedd to move pretrained GPT model to /home/works/TestData first")
class TestModularAudioGPTModel:
    @pytest.mark.unit
    def test_init_and_train(self, llm_model_config, perception_model_config, trainer_config):
        llm_model_config.model.pretrained_audio_model = "stt_en_fastconformer_transducer_large"
        llm_model_config.model.perception = perception_model_config
        trainer, llm_model_config.trainer = trainer_config
        model = ModularAudioGPTModel.restore_from_pretrained_models(llm_model_config, trainer=trainer)

        with tempfile.TemporaryDirectory() as tmpdir:
            save_path = str(Path(tmpdir) / "model.nemo")
            model.train()
            model.save_to(save_path)

    @pytest.mark.unit
    def test_prepare_llm_input(self, llm_model_config, perception_model_config, trainer_config, test_batch):
        llm_model_config.model.pretrained_audio_model = "stt_en_fastconformer_transducer_large"
        llm_model_config.model.perception = perception_model_config
        trainer, llm_model_config.trainer = trainer_config
        model = ModularAudioGPTModel.restore_from_pretrained_models(llm_model_config, trainer=trainer)
        model.cuda()
        model.train()
        batch = {key: val.cuda(non_blocking=True) for key, val in test_batch.items()}
        encoder_input, attention_mask, labels, loss_mask, encoder_length = model.prepare_llm_input(batch)
        assert encoder_input.shape == (17, 2, 768)
        assert np.allclose(encoder_input.sum().cpu().detach().numpy(), 15.783691)
        assert attention_mask.shape == (2, 1, 17, 17)
        assert labels.shape == (2, 17)
        assert np.allclose(loss_mask.sum(axis=1).cpu().numpy(), [2, 1])
        assert np.allclose(encoder_length.cpu().numpy(), (16, 15))

    @pytest.mark.unit
    def test_training_step(self, llm_model_config, perception_model_config, trainer_config, test_batch):
        llm_model_config.model.pretrained_audio_model = "stt_en_fastconformer_transducer_large"
        llm_model_config.model.perception = perception_model_config
        trainer, llm_model_config.trainer = trainer_config
        model = ModularAudioGPTModel.restore_from_pretrained_models(llm_model_config, trainer=trainer)
        model.cuda()
        model.on_train_start()
        model.setup()
        model.train()
        loss_mean = model.training_step(iter([test_batch]), None)
        assert np.allclose(loss_mean.cpu().detach().numpy(), 5.7052)

    @pytest.mark.unit
    def test_validation_step(self, llm_model_config, perception_model_config, trainer_config, test_batch):
        llm_model_config.model.pretrained_audio_model = "stt_en_fastconformer_transducer_large"
        llm_model_config.model.perception = perception_model_config
        trainer, llm_model_config.trainer = trainer_config
        model = ModularAudioGPTModel.restore_from_pretrained_models(llm_model_config, trainer=trainer)
        model.cuda()
        model.train()
        batch = {key: val.cuda(non_blocking=True) for key, val in test_batch.items()}
        loss_mean = model.validation_step(iter([batch]), 0)
        assert np.allclose(loss_mean['loss'].cpu().detach().numpy(), 5.7052)

    @pytest.mark.unit
    def test_predict_step(self, llm_model_config, perception_model_config, trainer_config, test_batch):
        llm_model_config.model.pretrained_audio_model = "stt_en_fastconformer_transducer_large"
        llm_model_config.model.perception = perception_model_config
        trainer, llm_model_config.trainer = trainer_config
        model = ModularAudioGPTModel.restore_from_pretrained_models(llm_model_config, trainer=trainer)
        model.cuda()
        model.train()
        batch = {key: val.cuda(non_blocking=True) for key, val in test_batch.items()}
        response = model.predict_step(batch, 0, 0)
        ground_truth = 'to suit you. Please note these are lecture notes from an alternate presentation. Copyright  ⁇ '
        assert response['sentences'][0] == ground_truth

    @pytest.mark.unit
    def test_concat_multi_features(self, llm_model_config, perception_model_config, trainer_config):
        llm_model_config.model.pretrained_audio_model = "stt_en_fastconformer_transducer_large"
        llm_model_config.model.perception = perception_model_config
        trainer, llm_model_config.trainer = trainer_config
        model = ModularAudioGPTModel.restore_from_pretrained_models(llm_model_config, trainer=trainer)
        model.eval()

        feat_dim = 32
        encoded = [torch.ones([3, 16, feat_dim]), torch.ones([3, 16, feat_dim])]
        encoded_len = [torch.LongTensor([12, 8, 4]), torch.LongTensor([12, 8, 4])]
        input_embeds = torch.zeros([2, 32, feat_dim])
        input_length = torch.LongTensor([32, 28])
        context_start_idx = [[0, 4, 12, 20], [0, 8, 16, 25]]
        encoder_input, encoder_length = model._concat_multi_features(
            encoded, encoded_len, input_embeds, input_length, context_start_idx
        )
        assert encoder_input.shape == (2, 56, feat_dim)  # max audio_len + text_len = (12 + 8 + 4) + 32 = 56
        assert encoder_length.shape == (2,)
        assert np.allclose(encoder_length.cpu().numpy(), (56, 52))
        assert encoder_input[0, : context_start_idx[0][1]].sum() == 0  # first 4 features are text features
        assert np.allclose(
            encoder_input[0, context_start_idx[0][1] : context_start_idx[0][1] + encoded_len[0][0]],
            torch.ones([encoded_len[0][0], feat_dim]),
        )

    @pytest.mark.unit
    def test_shift_tokens_by_multi_audios(self):
        """This test is put here because its functionality is similar to _concat_multi_features()"""
        encoder_max_length = 64
        audio_len = [torch.LongTensor([12, 8, 4]), torch.LongTensor([12, 8, 4])]
        context_tokens = torch.ones([2, 32])
        context_length = torch.LongTensor([32, 28])
        context_start_idx = [[0, 4, 12, 20], [0, 8, 16, 25]]
        new_context_tokens = shift_tokens_by_multi_audios(
            context_tokens, context_length, audio_len, context_start_idx, encoder_max_length
        )
        assert new_context_tokens.shape == (2, 64)
        assert np.allclose(new_context_tokens[0, : context_start_idx[0][1]], torch.ones([context_start_idx[0][1]]))
        assert np.allclose(
            new_context_tokens[0, context_start_idx[0][1] : context_start_idx[0][1] + audio_len[0][0]],
            torch.zeros([audio_len[0][0]]),
        )



# Copyright (c) 2022, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# The MIT License (MIT)
# Copyright (c) 2020, nicolas deutschmann
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import numpy as np
import torch
import torch.nn.functional as F


def piecewise_linear_transform(x, q_tilde, compute_jacobian=True, outlier_passthru=True):
    """Apply an element-wise piecewise-linear transformation to some variables

    Args:
        x : torch.Tensor
            a tensor with shape (N,k) where N is the batch dimension while k is the dimension of the variable space. This variable span the k-dimensional unit
        hypercube

        q_tilde: torch.Tensor
                is a tensor with shape (N,k,b) where b is the number of bins.
                This contains the un-normalized heights of the bins of the piecewise-constant PDF for dimension k,
                i.e. q_tilde lives in all of R and we don't impose a constraint on their sum yet.
                Normalization is imposed in this function using softmax.

        compute_jacobian : bool, optional
                            determines whether the jacobian should be compute or None is returned

    Returns:
        tuple of torch.Tensor
            pair `(y,h)`.
            - `y` is a tensor with shape (N,k) living in the k-dimensional unit hypercube
            - `j` is the jacobian of the transformation with shape (N,) if compute_jacobian==True, else None.
    """
    logj = None

    third_dimension_softmax = torch.nn.Softmax(dim=2)

    # Compute the bin width w
    N, k, b = q_tilde.shape
    Nx, kx = x.shape
    assert N == Nx and k == kx, "Shape mismatch"

    w = 1.0 / b

    # Compute normalized bin heights with softmax function on bin dimension
    q = 1.0 / w * third_dimension_softmax(q_tilde)
    # x is in the mx-th bin: x \in [0,1],
    # mx \in [[0,b-1]], so we clamp away the case x == 1
    mx = torch.clamp(torch.floor(b * x), 0, b - 1).to(torch.long)
    # Need special error handling because trying to index with mx
    # if it contains nans will lock the GPU. (device-side assert triggered)
    if torch.any(torch.isnan(mx)).item() or torch.any(mx < 0) or torch.any(mx >= b):
        raise AvertedCUDARuntimeError("NaN detected in PWLinear bin indexing")

    # We compute the output variable in-place
    out = x - mx * w  # alpha (element of [0.,w], the position of x in its bin

    # Multiply by the slope
    # q has shape (N,k,b), mxu = mx.unsqueeze(-1) has shape (N,k) with entries that are a b-index
    # gather defines slope[i, j, k] = q[i, j, mxu[i, j, k]] with k taking only 0 as a value
    # i.e. we say slope[i, j] = q[i, j, mx [i, j]]
    slopes = torch.gather(q, 2, mx.unsqueeze(-1)).squeeze(-1)
    out = out * slopes
    # The jacobian is the product of the slopes in all dimensions

    # Compute the integral over the left-bins.
    # 1. Compute all integrals: cumulative sum of bin height * bin weight.
    # We want that index i contains the cumsum *strictly to the left* so we shift by 1
    # leaving the first entry null, which is achieved with a roll and assignment
    q_left_integrals = torch.roll(torch.cumsum(q, 2) * w, 1, 2)
    q_left_integrals[:, :, 0] = 0

    # 2. Access the correct index to get the left integral of each point and add it to our transformation
    out = out + torch.gather(q_left_integrals, 2, mx.unsqueeze(-1)).squeeze(-1)

    # Regularization: points must be strictly within the unit hypercube
    # Use the dtype information from pytorch
    eps = torch.finfo(out.dtype).eps
    out = out.clamp(min=eps, max=1.0 - eps)
    oob_mask = torch.logical_or(x < 0.0, x > 1.0).detach().float()
    if outlier_passthru:
        out = out * (1 - oob_mask) + x * oob_mask
        slopes = slopes * (1 - oob_mask) + oob_mask

    if compute_jacobian:
        # logj = torch.log(torch.prod(slopes.float(), 1))
        logj = torch.sum(torch.log(slopes), 1)
    del slopes

    return out, logj


def piecewise_linear_inverse_transform(y, q_tilde, compute_jacobian=True, outlier_passthru=True):
    """
    Apply inverse of an element-wise piecewise-linear transformation to some
    variables

    Args:
        y : torch.Tensor
            a tensor with shape (N,k) where N is the batch dimension while k is the dimension of the variable space. This variable span the k-dimensional unit hypercube

        q_tilde: torch.Tensor
                is a tensor with shape (N,k,b) where b is the number of bins. This contains the un-normalized heights of the bins of the piecewise-constant PDF for dimension k, i.e. q_tilde lives in all of R and we don't impose a constraint on their sum yet. Normalization is imposed in this function using softmax.

        compute_jacobian : bool, optional
                            determines whether the jacobian should be compute or None is returned

    Returns:
        tuple of torch.Tensor
            pair `(x,h)`.
            - `x` is a tensor with shape (N,k) living in the k-dimensional unit hypercube
            - `j` is the jacobian of the transformation with shape (N,) if compute_jacobian==True, else None.
    """

    third_dimension_softmax = torch.nn.Softmax(dim=2)
    # Compute the bin width w
    N, k, b = q_tilde.shape
    Ny, ky = y.shape
    assert N == Ny and k == ky, "Shape mismatch"

    w = 1.0 / b

    # Compute normalized bin heights with softmax function on the bin dimension
    q = 1.0 / w * third_dimension_softmax(q_tilde)

    # Compute the integral over the left-bins in the forward transform.
    # 1. Compute all integrals: cumulative sum of bin height * bin weight.
    # We want that index i contains the cumsum *strictly to the left*,
    # so we shift by 1 leaving the first entry null,
    # which is achieved with a roll and assignment
    q_left_integrals = torch.roll(torch.cumsum(q.float(), 2) * w, 1, 2)
    q_left_integrals[:, :, 0] = 0

    # Find which bin each y belongs to by finding the smallest bin such that
    # y - q_left_integral is positive

    edges = (y.unsqueeze(-1) - q_left_integrals).detach()
    # y and q_left_integrals are between 0 and 1,
    # so that their difference is at most 1.
    # By setting the negative values to 2., we know that the
    # smallest value left is the smallest positive
    edges[edges < 0] = 2.0
    edges = torch.clamp(torch.argmin(edges, dim=2), 0, b - 1).to(torch.long)

    # Need special error handling because trying to index with mx
    # if it contains nans will lock the GPU. (device-side assert triggered)
    if torch.any(torch.isnan(edges)).item() or torch.any(edges < 0) or torch.any(edges >= b):
        raise AvertedCUDARuntimeError("NaN detected in PWLinear bin indexing")

    # Gather the left integrals at each edge. See comment about gathering in q_left_integrals
    # for the unsqueeze
    q_left_integrals = q_left_integrals.gather(2, edges.unsqueeze(-1)).squeeze(-1)

    # Gather the slope at each edge.
    q = q.gather(2, edges.unsqueeze(-1)).squeeze(-1)

    # Build the output
    x = (y - q_left_integrals) / q + edges * w

    # Regularization: points must be strictly within the unit hypercube
    # Use the dtype information from pytorch
    eps = torch.finfo(x.dtype).eps
    x = x.clamp(min=eps, max=1.0 - eps)
    oob_mask = torch.logical_or(y < 0.0, y > 1.0).detach().float()
    if outlier_passthru:
        x = x * (1 - oob_mask) + y * oob_mask
        q = q * (1 - oob_mask) + oob_mask

    # Prepare the jacobian
    logj = None
    if compute_jacobian:
        # logj = - torch.log(torch.prod(q, 1))
        logj = -torch.sum(torch.log(q.float()), 1)
    return x.detach(), logj


def unbounded_piecewise_quadratic_transform(x, w_tilde, v_tilde, upper=1, lower=0, inverse=False):
    assert upper > lower
    _range = upper - lower
    inside_interval_mask = (x >= lower) & (x < upper)
    outside_interval_mask = ~inside_interval_mask

    outputs = torch.zeros_like(x)
    log_j = torch.zeros_like(x)

    outputs[outside_interval_mask] = x[outside_interval_mask]
    log_j[outside_interval_mask] = 0

    output, _log_j = piecewise_quadratic_transform(
        (x[inside_interval_mask] - lower) / _range,
        w_tilde[inside_interval_mask, :],
        v_tilde[inside_interval_mask, :],
        inverse=inverse,
    )
    outputs[inside_interval_mask] = output * _range + lower
    if not inverse:
        # the before and after transformation cancel out, so the log_j would be just as it is.
        log_j[inside_interval_mask] = _log_j
    else:
        log_j = None
    return outputs, log_j


def weighted_softmax(v, w):
    # to avoid NaN...
    v = v - torch.max(v, dim=-1, keepdim=True)[0]
    v = torch.exp(v) + 1e-8  # to avoid NaN...
    v_sum = torch.sum((v[..., :-1] + v[..., 1:]) / 2 * w, dim=-1, keepdim=True)
    return v / v_sum


def piecewise_quadratic_transform(x, w_tilde, v_tilde, inverse=False):
    """Element-wise piecewise-quadratic transformation
    Args:
        x : torch.Tensor
            *, The variable spans the D-dim unit hypercube ([0,1))
        w_tilde : torch.Tensor
            * x K defined in the paper
        v_tilde : torch.Tensor
            * x (K+1) defined in the paper
        inverse : bool
            forward or inverse
    Returns:
        c : torch.Tensor
            *, transformed value
        log_j : torch.Tensor
            *, log determinant of the Jacobian matrix
    """
    w = torch.softmax(w_tilde, dim=-1)
    v = weighted_softmax(v_tilde, w)
    w_cumsum = torch.cumsum(w, dim=-1)
    # force sum = 1
    w_cumsum[..., -1] = 1.0
    w_cumsum_shift = F.pad(w_cumsum, (1, 0), 'constant', 0)
    cdf = torch.cumsum((v[..., 1:] + v[..., :-1]) / 2 * w, dim=-1)
    # force sum = 1
    cdf[..., -1] = 1.0
    cdf_shift = F.pad(cdf, (1, 0), 'constant', 0)

    if not inverse:
        # * x D x 1, (w_cumsum[idx-1] < x <= w_cumsum[idx])
        bin_index = torch.searchsorted(w_cumsum, x.unsqueeze(-1))
    else:
        # * x D x 1, (cdf[idx-1] < x <= cdf[idx])
        bin_index = torch.searchsorted(cdf, x.unsqueeze(-1))

    w_b = torch.gather(w, -1, bin_index).squeeze(-1)
    w_bn1 = torch.gather(w_cumsum_shift, -1, bin_index).squeeze(-1)
    v_b = torch.gather(v, -1, bin_index).squeeze(-1)
    v_bp1 = torch.gather(v, -1, bin_index + 1).squeeze(-1)
    cdf_bn1 = torch.gather(cdf_shift, -1, bin_index).squeeze(-1)

    if not inverse:
        alpha = (x - w_bn1) / w_b.clamp(min=torch.finfo(w_b.dtype).eps)
        c = (alpha ** 2) / 2 * (v_bp1 - v_b) * w_b + alpha * v_b * w_b + cdf_bn1

        # just sum of log pdfs
        log_j = torch.lerp(v_b, v_bp1, alpha).clamp(min=torch.finfo(c.dtype).eps).log()

        # make sure it falls into [0,1)
        c = c.clamp(min=torch.finfo(c.dtype).eps, max=1.0 - torch.finfo(c.dtype).eps)
        return c, log_j
    else:
        # quadratic equation for alpha
        # alpha should fall into (0, 1]. Since a, b > 0, the symmetry axis -b/2a < 0 and we should pick the larger root
        # skip calculating the log_j in inverse since we don't need it
        a = (v_bp1 - v_b) * w_b / 2
        b = v_b * w_b
        c = cdf_bn1 - x
        alpha = (-b + torch.sqrt((b ** 2) - 4 * a * c)) / (2 * a)
        inv = alpha * w_b + w_bn1

        # make sure it falls into [0,1)
        inv = inv.clamp(min=torch.finfo(c.dtype).eps, max=1.0 - torch.finfo(inv.dtype).eps)
        return inv, None


def piecewise_rational_quadratic_transform(
    inputs,
    unnormalized_widths,
    unnormalized_heights,
    unnormalized_derivatives,
    inverse=False,
    tails=None,
    tail_bound=1.0,
    min_bin_width=1e-3,
    min_bin_height=1e-3,
    min_derivative=1e-3,
):

    if tails is None:
        spline_fn = rational_quadratic_spline
        spline_kwargs = {}
    else:
        spline_fn = unconstrained_rational_quadratic_spline
        spline_kwargs = {'tails': tails, 'tail_bound': tail_bound}

    outputs, logabsdet = spline_fn(
        inputs=inputs,
        unnormalized_widths=unnormalized_widths,
        unnormalized_heights=unnormalized_heights,
        unnormalized_derivatives=unnormalized_derivatives,
        inverse=inverse,
        min_bin_width=min_bin_width,
        min_bin_height=min_bin_height,
        min_derivative=min_derivative,
        **spline_kwargs
    )
    return outputs, logabsdet


def searchsorted(bin_locations, inputs, eps=1e-6):
    bin_locations[..., -1] += eps
    return torch.sum(inputs[..., None] >= bin_locations, dim=-1) - 1


def unconstrained_rational_quadratic_spline(
    inputs,
    unnormalized_widths,
    unnormalized_heights,
    unnormalized_derivatives,
    inverse=False,
    tails='linear',
    tail_bound=1.0,
    min_bin_width=1e-3,
    min_bin_height=1e-3,
    min_derivative=1e-3,
):
    inside_interval_mask = (inputs >= -tail_bound) & (inputs <= tail_bound)
    outside_interval_mask = ~inside_interval_mask

    outputs = torch.zeros_like(inputs)
    logabsdet = torch.zeros_like(inputs)

    if tails == 'linear':
        unnormalized_derivatives = F.pad(unnormalized_derivatives, pad=(1, 1))
        constant = np.log(np.exp(1 - min_derivative) - 1)
        unnormalized_derivatives[..., 0] = constant
        unnormalized_derivatives[..., -1] = constant

        outputs[outside_interval_mask] = inputs[outside_interval_mask]
        logabsdet[outside_interval_mask] = 0
    else:
        raise RuntimeError('{} tails are not implemented.'.format(tails))

    outputs[inside_interval_mask], logabsdet[inside_interval_mask] = rational_quadratic_spline(
        inputs=inputs[inside_interval_mask],
        unnormalized_widths=unnormalized_widths[inside_interval_mask, :],
        unnormalized_heights=unnormalized_heights[inside_interval_mask, :],
        unnormalized_derivatives=unnormalized_derivatives[inside_interval_mask, :],
        inverse=inverse,
        left=-tail_bound,
        right=tail_bound,
        bottom=-tail_bound,
        top=tail_bound,
        min_bin_width=min_bin_width,
        min_bin_height=min_bin_height,
        min_derivative=min_derivative,
    )

    return outputs, logabsdet


def rational_quadratic_spline(
    inputs,
    unnormalized_widths,
    unnormalized_heights,
    unnormalized_derivatives,
    inverse=False,
    left=0.0,
    right=1.0,
    bottom=0.0,
    top=1.0,
    min_bin_width=1e-3,
    min_bin_height=1e-3,
    min_derivative=1e-3,
):

    if torch.min(inputs) < left or torch.max(inputs) > right:
        raise ValueError('Input to a transform is not within its domain')

    num_bins = unnormalized_widths.shape[-1]

    if min_bin_width * num_bins > 1.0:
        raise ValueError('Minimal bin width too large for the number of bins')
    if min_bin_height * num_bins > 1.0:
        raise ValueError('Minimal bin height too large for the number of bins')

    widths = F.softmax(unnormalized_widths, dim=-1)
    widths = min_bin_width + (1 - min_bin_width * num_bins) * widths
    cumwidths = torch.cumsum(widths, dim=-1)
    cumwidths = F.pad(cumwidths, pad=(1, 0), mode='constant', value=0.0)
    cumwidths = (right - left) * cumwidths + left
    cumwidths[..., 0] = left
    cumwidths[..., -1] = right
    widths = cumwidths[..., 1:] - cumwidths[..., :-1]

    derivatives = min_derivative + F.softplus(unnormalized_derivatives)

    heights = F.softmax(unnormalized_heights, dim=-1)
    heights = min_bin_height + (1 - min_bin_height * num_bins) * heights
    cumheights = torch.cumsum(heights, dim=-1)
    cumheights = F.pad(cumheights, pad=(1, 0), mode='constant', value=0.0)
    cumheights = (top - bottom) * cumheights + bottom
    cumheights[..., 0] = bottom
    cumheights[..., -1] = top
    heights = cumheights[..., 1:] - cumheights[..., :-1]

    if inverse:
        bin_idx = searchsorted(cumheights, inputs)[..., None]
    else:
        bin_idx = searchsorted(cumwidths, inputs)[..., None]

    input_cumwidths = cumwidths.gather(-1, bin_idx)[..., 0]
    input_bin_widths = widths.gather(-1, bin_idx)[..., 0]

    input_cumheights = cumheights.gather(-1, bin_idx)[..., 0]
    delta = heights / widths
    input_delta = delta.gather(-1, bin_idx)[..., 0]

    input_derivatives = derivatives.gather(-1, bin_idx)[..., 0]
    input_derivatives_plus_one = derivatives[..., 1:].gather(-1, bin_idx)[..., 0]

    input_heights = heights.gather(-1, bin_idx)[..., 0]

    if inverse:
        a = (inputs - input_cumheights) * (
            input_derivatives + input_derivatives_plus_one - 2 * input_delta
        ) + input_heights * (input_delta - input_derivatives)
        b = input_heights * input_derivatives - (inputs - input_cumheights) * (
            input_derivatives + input_derivatives_plus_one - 2 * input_delta
        )
        c = -input_delta * (inputs - input_cumheights)

        discriminant = b.pow(2) - 4 * a * c
        assert (discriminant >= 0).all()

        root = (2 * c) / (-b - torch.sqrt(discriminant))
        outputs = root * input_bin_widths + input_cumwidths

        theta_one_minus_theta = root * (1 - root)
        denominator = input_delta + (
            (input_derivatives + input_derivatives_plus_one - 2 * input_delta) * theta_one_minus_theta
        )
        derivative_numerator = input_delta.pow(2) * (
            input_derivatives_plus_one * root.pow(2)
            + 2 * input_delta * theta_one_minus_theta
            + input_derivatives * (1 - root).pow(2)
        )
        logabsdet = torch.log(derivative_numerator) - 2 * torch.log(denominator)

        return outputs, -logabsdet
    else:
        theta = (inputs - input_cumwidths) / input_bin_widths
        theta_one_minus_theta = theta * (1 - theta)

        numerator = input_heights * (input_delta * theta.pow(2) + input_derivatives * theta_one_minus_theta)
        denominator = input_delta + (
            (input_derivatives + input_derivatives_plus_one - 2 * input_delta) * theta_one_minus_theta
        )
        outputs = input_cumheights + numerator / denominator

        derivative_numerator = input_delta.pow(2) * (
            input_derivatives_plus_one * theta.pow(2)
            + 2 * input_delta * theta_one_minus_theta
            + input_derivatives * (1 - theta).pow(2)
        )
        logabsdet = torch.log(derivative_numerator) - 2 * torch.log(denominator)

        return outputs, logabsdet



# Copyright (c) 2022, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
from typing import Dict, List, Tuple

import numpy as np
from numpy import ndarray
from sklearn.preprocessing import PowerTransformer, QuantileTransformer, RobustScaler

from nemo.utils import logging

__all__ = ["IntCode", "FloatCode", "CategoryCode", "ColumnCodes"]


class Code(object):
    def compute_code(self, data_series: ndarray):
        """
        @params:
            data_series: an array of input data used to calculate mapping
        """
        raise NotImplementedError()

    def __init__(self, col_name: str, code_len: int, start_id: int, fillall: bool = True, hasnan: bool = True):
        """
        @params:
            col_name: name of the column
            code_len: number of tokens used to code the column.
            start_id: offset for token_id. 
            fillall: if True, reserve space for digit number even the digit number is
            not present in the data_series. Otherwise, only reserve space for the numbers
            in the data_series. 
            hasnan: if True, reserve space for nan
        """
        self.name = col_name
        self.code_len = code_len
        self.start_id = start_id
        self.end_id = start_id
        self.fillall = fillall
        self.hasnan = hasnan

    def encode(self, item: str) -> List[int]:
        raise NotImplementedError()

    def decode(self, ids: List[int]) -> str:
        raise NotImplementedError()

    @property
    def code_range(self) -> List[Tuple[int, int]]:
        """
        get the vocab id range for each of the encoded tokens
        @returns [(min, max), (min, max), ...]
        """
        return [(self.start_id, self.end_id)]


class IntCode(Code):
    def __init__(
        self, col_name: str, code_len: int, start_id: int, fillall: bool = True, base: int = 100, hasnan: bool = True
    ):
        super().__init__(col_name, code_len, start_id, fillall, hasnan)
        self.base = base
        self.int_min: int = None

    def compute_code(self, data_series: ndarray):
        significant_val = self.array_convert_to_int(data_series)

        digits_id_to_item = [{} for _ in range(self.code_len)]
        digits_item_to_id = [{} for _ in range(self.code_len)]
        for i in range(self.code_len):
            id_to_item = digits_id_to_item[i]
            item_to_id = digits_item_to_id[i]
            v = (significant_val // self.base ** i) % self.base
            if self.fillall:
                uniq_items = range(0, self.base)
            else:
                uniq_items = sorted(np.unique(v).tolist())
            for k in range(len(uniq_items)):
                item = str(uniq_items[k])
                item_to_id[item] = self.end_id
                id_to_item[self.end_id] = item
                self.end_id += 1
        self.digits_id_to_item = digits_id_to_item
        self.digits_item_to_id = digits_item_to_id
        self.NA_token = 'nan'
        if self.hasnan:
            self.end_id += 1  # add the N/A token
            codes = []
            ranges = self.code_range
            for i in ranges:
                codes.append(i[1] - 1)
            self.NA_token_id = codes

    def array_convert_to_int(self, val: ndarray):
        val = val.astype(int)
        self.int_min = val.min()
        return val - self.int_min

    def convert_to_int(self, val: float) -> int:
        return int(val) - self.int_min

    def reverse_convert_to_int(self, val: int) -> int:
        return val + self.int_min

    @property
    def code_range(self) -> List[Tuple[int, int]]:
        """
        get the vocab id range for each of the encoded tokens
        @returns [(min, max), (min, max), ...]
        """
        # first largest digits
        outputs = []
        c = 0
        for i in reversed(range(self.code_len)):
            ids = self.digits_id_to_item[i].keys()
            if c == 0:
                if self.hasnan:
                    outputs.append((min(ids), max(ids) + 2))  # the first token contains the N/A
                else:
                    outputs.append((min(ids), max(ids) + 1))  # non N/A
            else:
                outputs.append((min(ids), max(ids) + 1))
            c += 1
        return outputs

    def encode(self, item: str) -> List[int]:
        if self.hasnan and item == self.NA_token:
            return self.NA_token_id
        elif not self.hasnan and item == self.NA_token:
            raise ValueError(f"colum {self.name} cannot handle nan, please set hasnan=True")
        val = float(item)
        val_int = self.convert_to_int(val)
        digits = []
        for i in range(self.code_len):
            digit = (val_int // self.base ** i) % self.base
            digits.append(str(digit))
        if (val_int // self.base ** self.code_len) != 0:
            raise ValueError("not right length")
        codes = []
        for i in reversed(range(self.code_len)):
            digit_str = digits[i]
            if digit_str in self.digits_item_to_id[i]:
                codes.append(self.digits_item_to_id[i][digit_str])
            else:
                # find the nearest encode id
                allowed_digits = np.array([int(d) for d in self.digits_item_to_id[i].keys()])
                near_id = np.argmin(np.abs(allowed_digits - int(digit_str)))
                digit_str = str(allowed_digits[near_id])
                codes.append(self.digits_item_to_id[i][digit_str])
                logging.warning('out of domain num is encounterd, use nearest code')
        return codes

    def decode(self, ids: List[int]) -> str:
        if self.hasnan and ids[0] == self.NA_token_id[0]:
            return self.NA_token
        v = 0
        for i in reversed(range(self.code_len)):
            digit = int(self.digits_id_to_item[i][ids[self.code_len - i - 1]])
            v += digit * self.base ** i
        v = self.reverse_convert_to_int(v)
        return str(v)


class FloatCode(IntCode):
    def __init__(
        self,
        col_name: str,
        code_len: int,
        start_id: int,
        fillall: bool = True,
        base: int = 100,
        hasnan: bool = True,
        transform: str = 'quantile',
    ):
        super().__init__(col_name, code_len, start_id, fillall, base, hasnan)
        if transform == 'yeo-johnson':
            self.scaler = PowerTransformer(standardize=True)
        elif transform == 'quantile':
            self.scaler = QuantileTransformer(output_distribution='uniform', n_quantiles=100)
        elif transform == 'robust':
            self.scaler = RobustScaler()
        else:
            raise ValueError('Supported data transformations are "yeo-johnson", "quantile", and "robust"')

    def convert_to_int(self, val: float) -> int:
        val = np.expand_dims(np.array(val), axis=0)
        values = self.scaler.transform(val[:, None])[:, 0] - self.mval
        values = (values * self.base ** self.extra_digits).astype(int)
        output = values[0]
        return output

    def array_convert_to_int(self, val: ndarray):
        values = self.scaler.fit_transform(val[:, None])[:, 0]
        self.mval = values.min()
        values = values - self.mval
        digits = int(math.log(values.max(), self.base)) + 1
        # extra digits used for 'float' part of the number
        extra_digits = self.code_len - digits
        if extra_digits < 0:
            raise ValueError("need large length to code the nummber")
        self.extra_digits = extra_digits
        values = (values * self.base ** self.extra_digits).astype(int)
        return values

    def reverse_convert_to_int(self, val: int) -> float:
        val = val / self.base ** self.extra_digits
        val = np.expand_dims(np.array(val), axis=0)
        v = self.scaler.inverse_transform(val[:, None] + self.mval)[0, 0]
        return v

    def decode(self, ids: List[int]) -> str:
        if self.hasnan and ids[0] == self.NA_token_id[0]:
            return self.NA_token
        v = 0
        for i in reversed(range(self.code_len)):
            digit = int(self.digits_id_to_item[i][ids[self.code_len - i - 1]])
            v += digit * self.base ** i
        v = self.reverse_convert_to_int(v)
        accuracy = max(int(abs(np.log10(0.1 / self.base ** self.extra_digits))), 1)
        return f"{v:.{accuracy}f}"


class CategoryCode(Code):
    def __init__(self, col_name: str, start_id: int):
        super().__init__(col_name, 1, start_id, True, False)

    def compute_code(self, data_series: ndarray):
        uniq_items = np.unique(data_series).tolist()
        id_to_item = {}
        item_to_id = {}
        for i in range(len(uniq_items)):
            item = str(uniq_items[i])
            item_to_id[item] = self.end_id
            id_to_item[self.end_id] = item
            self.end_id += 1
        self.id_to_item = id_to_item
        self.item_to_id = item_to_id

    def encode(self, item) -> List[int]:
        return [self.item_to_id[item]]

    def decode(self, ids: List[int]) -> str:
        return self.id_to_item[ids[0]]


column_map = {"int": IntCode, "float": FloatCode, "category": CategoryCode}


class ColumnCodes(object):
    def __init__(self):
        self.column_codes: Dict[str, Code] = {}
        self.columns = []
        self.sizes = []

    @property
    def vocab_size(self):
        return self.column_codes[self.columns[-1]].end_id

    def register(self, name: str, ccode: Code):
        self.columns.append(name)
        self.column_codes[name] = ccode
        self.sizes.append(ccode.code_len)

    def encode(self, col: str, item: str) -> List[int]:
        if col in self.column_codes:
            return self.column_codes[col].encode(item)
        else:
            raise ValueError(f"cannot encode {col} {item}")

    def decode(self, col: str, ids: List[int]) -> str:
        if col in self.column_codes:
            return self.column_codes[col].decode(ids)
        else:
            raise ValueError("cannot decode")

    def get_range(self, column_id: int) -> List[Tuple[int, int]]:
        return self.column_codes[self.columns[column_id]].code_range

    @classmethod
    def get_column_codes(cls, column_configs, example_arrays):
        column_codes = cls()
        beg = 0
        cc = None
        for config in column_configs:
            col_name = config['name']
            coder = column_map[config['code_type']]
            args = config.get('args', {})
            start_id = beg if cc is None else cc.end_id
            args['start_id'] = start_id
            args['col_name'] = col_name
            cc = coder(**args)
            cc.compute_code(example_arrays[col_name])
            column_codes.register(col_name, cc)
        return column_codes



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import re
from functools import partial
from typing import TYPE_CHECKING, Callable

from omegaconf import DictConfig

from nemo.collections.asr.inference.streaming.state.state import StreamingState
from nemo.collections.asr.inference.utils.constants import POST_WORD_PUNCTUATION
from nemo.collections.asr.inference.utils.pipeline_utils import (
    get_leading_punctuation_regex_pattern,
    get_repeated_punctuation_regex_pattern,
)
from nemo.collections.asr.inference.utils.text_segment import Word, normalize_segments_inplace

if TYPE_CHECKING:
    from nemo.collections.asr.inference.itn.inverse_normalizer import AlignmentPreservingInverseNormalizer


class StreamingTextProcessor:
    """
    A streaming text post-processing module that applies punctuation & capitalization (PnC) and
    inverse text normalization (ITN) to ASR transcriptions in real-time.

    This class supports configurable pipelines where PnC and ITN can be enabled/disabled dynamically.
    It ensures that the final output adheres to proper punctuation, capitalization, and normalized text.
    """

    def __init__(
        self,
        itn_cfg: DictConfig,
        itn_model: AlignmentPreservingInverseNormalizer | None,
        asr_supported_puncts: set,
        asr_supports_punctuation: bool,
        confidence_aggregator: Callable,
        sep: str,
        enable_pnc: bool = False,
        enable_itn: bool = False,
    ):
        """
        Initialize the streaming text processor.

        Args:
            itn_cfg (DictConfig): ITN parameters.
            itn_model (AlignmentPreservingInverseNormalizer | None): Model for inverse text normalization (ITN).
            asr_supported_puncts (set): Set of punctuation marks recognized by the ASR model.
            asr_supports_punctuation (bool): Boolean indicating if the ASR model outputs punctuation.
            confidence_aggregator (Callable): Function for aggregating confidence scores.
            sep (str): String separator used in ASR output processing.
            enable_pnc (bool): Boolean to enable PnC. Default is False.
            enable_itn (bool): Boolean to enable ITN. Default is False.
        """

        self.pnc_enabled = enable_pnc and asr_supports_punctuation
        self.supports_punctuation = asr_supports_punctuation

        self.itn_model = itn_model
        self.itn_enabled = False
        if enable_itn:
            self.itn_enabled = itn_model is not None

        self.itn_runtime_params = {
            "batch_size": itn_cfg.batch_size,
            "n_jobs": itn_cfg.n_jobs,
        }
        self.itn_left_padding_size = itn_cfg.left_padding_size

        self.asr_supported_puncts = asr_supported_puncts
        self.asr_supported_puncts_str = ''.join(self.asr_supported_puncts)
        self.sep = sep
        self.rm_punctuation_capitalization_from_segments_fn = partial(
            normalize_segments_inplace, punct_marks=self.asr_supported_puncts, sep=self.sep
        )

        puncts_to_process = self.asr_supported_puncts
        self.leading_punctuation_regex_pattern = get_leading_punctuation_regex_pattern(puncts_to_process)
        self.repeated_punctuation_regex_pattern = get_repeated_punctuation_regex_pattern(puncts_to_process)

        self.alignment_aware_itn_model = None
        if self.itn_enabled:
            from nemo.collections.asr.inference.itn.batch_inverse_normalizer import (
                BatchAlignmentPreservingInverseNormalizer,
            )

            self.alignment_aware_itn_model = BatchAlignmentPreservingInverseNormalizer(
                itn_model=self.itn_model,
                sep=self.sep,
                asr_supported_puncts=self.asr_supported_puncts,
                post_word_punctuation=POST_WORD_PUNCTUATION,
                conf_aggregate_fn=confidence_aggregator,
            )

    def is_itn_enabled(self) -> bool:
        """Check if ITN is enabled"""
        return self.itn_enabled

    def is_pnc_enabled(self) -> bool:
        """Check if PnC is enabled"""
        return self.pnc_enabled

    def is_enabled(self) -> bool:
        """Check if PnC or ITN is enabled"""
        return self.is_pnc_enabled() or self.is_itn_enabled()

    def process(self, states: list[StreamingState]) -> None:
        """
        Apply PnC and ITN on the states.
        Args:
            states: (list[StreamingState]) List of StreamingState objects
        """
        word_boundary_states, segment_boundary_states = [], []
        for state in states:
            if state.options.is_word_level_output():
                word_boundary_states.append(state)
            else:
                segment_boundary_states.append(state)

        # Process states with word boundaries
        if word_boundary_states:
            self.process_states_with_word_boundaries(word_boundary_states)

        # Process states with segment boundaries
        if segment_boundary_states:
            self.process_states_with_segment_boundaries(segment_boundary_states)

        # Generate final transcript
        self.generate_final_transcript(word_boundary_states, segment_boundary_states)

    def process_states_with_segment_boundaries(self, states: list[StreamingState]) -> None:
        """
        Process states with segment boundaries.
        Args:
            states (list[StreamingState]): List of StreamingState objects that have segments
        """
        states_with_text = [state for state in states if len(state.segments) > 0]
        if len(states_with_text) == 0:
            return

        # if PnC & ITN DISABLED globally, remove PnC from the words if ASR supports punctuation
        if not self.is_enabled():
            if self.supports_punctuation:
                segments = []
                for state in states_with_text:
                    for i, seg in enumerate(state.segments):
                        if not state.processed_segment_mask[i]:
                            segments.append(seg)
                            state.processed_segment_mask[i] = True
                self.rm_punctuation_capitalization_from_segments_fn(segments)
            return

        # Remove PnC from states where PnC is disabled
        for state in states_with_text:
            if (not state.options.enable_pnc) or (not self.is_pnc_enabled()):
                if self.supports_punctuation:
                    self.rm_punctuation_capitalization_from_segments_fn(state.segments)

        # Apply ITN
        if self.is_itn_enabled():  # If ITN ENABLED globally
            # collect texts
            texts = []
            for i, state in enumerate(states_with_text):
                # if ITN is disabled for this state
                if not state.options.enable_itn:
                    continue

                for j, seg in enumerate(state.segments):
                    if state.processed_segment_mask[j]:  # if the segment is already processed, skip it
                        continue
                    texts.append((i, j, seg.text))

            if len(texts) > 0:
                # apply ITN
                processed_texts = self.itn_model.inverse_normalize_list(
                    texts=[text for _, _, text in texts], params=self.itn_runtime_params
                )
                # update states with ITN-processed texts
                for (i, j, _), processed_text in zip(texts, processed_texts):
                    states_with_text[i].segments[j].text = processed_text

        # --> Apply External PnC here (if needed)

        # mark all segments as processed
        for state in states_with_text:
            if state.options.enable_pnc:
                for seg in state.segments:
                    if self.leading_punctuation_regex_pattern:
                        seg.text = re.sub(self.leading_punctuation_regex_pattern, r'\1', seg.text)
                    if self.repeated_punctuation_regex_pattern:
                        seg.text = re.sub(self.repeated_punctuation_regex_pattern, r'\1', seg.text)
            state.processed_segment_mask = [True] * len(state.segments)

    def process_states_with_word_boundaries(self, states: list[StreamingState]) -> None:
        """
        Apply PnC and ITN on the states.
        Args:
            states: (list[StreamingState]) List of StreamingState objects
        """
        # Get the indices of the states that have new words to process
        indices, asr_words_list = self.prepare_asr_words(states)

        # If PnC & ITN DISABLED globally, remove PnC from the words
        # Does not matter that individual request has enabled itn or pnc
        if not self.is_enabled():
            self.handle_plain_asr_transcriptions(states, indices, asr_words_list)
            return

        # Keep or remove PnC from the words
        for idx, jdx, z in indices:
            if not states[idx].options.enable_pnc and self.supports_punctuation:
                self.rm_punctuation_capitalization_from_segments_fn(asr_words_list[jdx])
            states[idx].pnc_words[-z:] = asr_words_list[jdx][-z:]

        # If ITN is disabled globally, do nothing
        if not self.itn_enabled:
            return

        # Apply Inverse Text Normalization (ITN)
        self.apply_itn(states, indices)
        self.realign_punctuated_words(states, indices)

    def realign_punctuated_words(self, states: list[StreamingState], indices: list[tuple]) -> None:
        """
        Realign punctuation and capitalization after applying ITN.
        Ensures that capitalization and punctuation marks from the original ASR output
        are properly reflected in the final ITN-processed text.

        Args:
            states (list[StreamingState]): List of StreamingState objects to be updated.
            indices (list[tuple]): Indices of words within states that need realignment.
        """
        for idx, _, z in indices:
            state = states[idx]
            if not state.options.enable_itn:
                continue

            z_idx = len(state.words) - z

            itn_idx = len(state.itn_words)
            for sids, _, _ in reversed(state.word_alignment):
                st, et = sids[0], sids[-1]
                itn_idx -= 1
                if st < z_idx and et < z_idx:
                    break

                last_char = state.pnc_words[et].text[-1]
                first_char = state.pnc_words[st].text[0]

                itn_word_orig = state.itn_words[itn_idx]
                itn_word_copy = itn_word_orig.copy()
                itn_word_text = itn_word_copy.text.lower()

                # preserve the first char capitalization
                first_word = state.pnc_words[st].copy()
                first_char_is_upper = first_word.text[0].isupper()
                first_word.normalize_text_inplace(self.asr_supported_puncts, self.sep)
                if first_char_is_upper and itn_word_text.startswith(first_word.text):
                    itn_word_orig.capitalize()

                # preserve the last punctuation mark
                if last_char in self.asr_supported_puncts:
                    itn_word_orig.text = itn_word_orig.text.rstrip(self.asr_supported_puncts_str) + last_char

                # preserve the first punctuation mark
                if first_char in self.asr_supported_puncts:
                    itn_word_orig.text = first_char + itn_word_orig.text.lstrip(self.asr_supported_puncts_str)

                state.itn_words[itn_idx] = itn_word_orig

    def prepare_asr_words(self, states: list[StreamingState]) -> tuple[list[tuple], list[list[Word]]]:
        """
        Find the indices of the states that have words to process.
        Args:
            states: (list[StreamingState]) List of StreamingState objects
        Returns:
            tuple[list[tuple], list[list[Word]]]:
                indices: list of indices of the states that have words to process
                asr_words_list: list of words to process
        """
        indices, asr_words_list = [], []

        jdx = 0
        for idx, state in enumerate(states):
            if (n_not_punctuated_words := len(state.words) - len(state.pnc_words)) == 0:
                continue

            words_list = [word.copy() for word in state.words[-n_not_punctuated_words:]]
            asr_words_list.append(words_list)
            state.pnc_words.extend([None] * n_not_punctuated_words)
            indices.append((idx, jdx, len(words_list)))
            jdx += 1

        return indices, asr_words_list

    def handle_plain_asr_transcriptions(
        self, states: list[StreamingState], indices: list[tuple], asr_words_list: list[list[Word]]
    ) -> None:
        """
        Handle scenarios where PnC and ITN are disabled.
        In such cases, remove Punctuation and Capitalization from the words.
        Args:
            states: (list[StreamingState]) List of StreamingState objects
            indices: (list[tuple]) List of indices of the states that have words to process
            asr_words_list: (list[list[Word]]) List of words
        """
        if self.supports_punctuation:
            self.rm_punctuation_capitalization_from_segments_fn(asr_words_list)

        for idx, jdx, z in indices:
            states[idx].pnc_words[-z:] = asr_words_list[jdx][-z:]

    def apply_itn(self, states: list[StreamingState], indices: list[tuple]) -> None:
        """
        Apply Inverse Text Normalization (ITN) on the states.
        Calculates the lookback for ITN and updates the states with the ITN results.
        Args:
            states: (list[StreamingState]) List of StreamingState objects
            indices: (list[tuple]) List of indices of the states that have words to process
        """
        itn_indices, asr_words_list, pnc_words_list = [], [], []
        jdx = 0
        for state_idx, _, _ in indices:
            state = states[state_idx]
            if not state.options.enable_itn:
                continue
            s, t, cut_point = self.calculate_itn_lookback(state)
            asr_words_list.append([word.copy() for word in state.words[s:]])
            pnc_words_list.append([word.copy() for word in state.pnc_words[s:]])
            itn_indices.append((state_idx, jdx, s, t, cut_point))
            jdx += 1
        output = self.alignment_aware_itn_model(
            asr_words_list, pnc_words_list, self.itn_runtime_params, return_alignment=True
        )
        self.update_itn_words(states, output, itn_indices)

    def calculate_itn_lookback(self, state: StreamingState) -> tuple[int, int, int]:
        """
        Calculate the lookback for ITN.
        Args:
            state: (StreamingState) StreamingState object
        Returns:
            Start index (int): Start index of the source (non itn-ed) words
            Target index (int): Start index of the target (itn-ed) words
            Cut point (int): Index to cut the source words
        """
        s, t, cut_point = 0, 0, len(state.itn_words)
        word_alignment = list(reversed(state.word_alignment))
        for idx, (sidx, tidx, _) in enumerate(word_alignment, start=1):
            s, t = sidx[0], tidx[0]
            state.word_alignment.pop()
            cut_point -= 1
            if idx == self.itn_left_padding_size:
                break
        return s, t, cut_point

    @staticmethod
    def update_itn_words(states: list[StreamingState], output: list[tuple], indices: list[tuple]) -> None:
        """
        Update the states with the ITN results.
        Updates the word_alignment and itn_words in the states.
        Args:
            states: (list[StreamingState]) List of StreamingState objects
            output: (list[tuple]) List of output tuples containing the spans and alignment
            indices: (list[tuple]) List of indices of the states that have words to process
        """
        for state_idx, jdx, s, t, cut_point in indices:
            state = states[state_idx]
            spans, alignment = output[jdx]
            for sidx, tidx, sclass in alignment:
                sidx = [k + s for k in sidx]
                tidx = [k + t for k in tidx]
                state.word_alignment.append((sidx, tidx, sclass))

            state.itn_words = state.itn_words[:cut_point] + spans
            assert len(state.word_alignment) == len(state.itn_words)

    def generate_final_transcript(
        self, word_boundary_states: list[StreamingState], segment_boundary_states: list[StreamingState]
    ) -> None:
        """
        Generate final transcript based on enabled features and word count.
        Args:
            word_boundary_states (list[StreamingState]): The streaming state containing words
            segment_boundary_states (list[StreamingState]): The streaming state containing segments
        """
        # Generate final transcript for word boundary states
        for state in word_boundary_states:
            attr_name = "itn_words" if state.options.enable_itn else "pnc_words"
            words = getattr(state, attr_name)
            for word in words:
                state.final_segments.append(word.copy())
                state.final_transcript += word.text + self.sep
            state.final_transcript = state.final_transcript.rstrip(self.sep)

        # Generate final transcript for segment boundary states
        for state in segment_boundary_states:
            for segment in state.segments:
                state.final_segments.append(segment.copy())
                state.final_transcript += segment.text + self.sep
            state.final_transcript = state.final_transcript.rstrip(self.sep)



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

import lightning.pytorch as pl
import nemo_run as run
import torch
from megatron.core.distributed import DistributedDataParallelConfig
from megatron.core.optimizer import OptimizerConfig
from transformers.models.qwen2_vl.image_processing_qwen2_vl import Qwen2VLImageProcessor

from nemo import lightning as nl
from nemo.collections import llm, vlm
from nemo.collections.common.tokenizers.huggingface.auto_tokenizer import AutoTokenizer
from nemo.collections.llm.recipes.log.default import tensorboard_logger
from nemo.collections.vlm.qwen2vl.model import Qwen25VLVisionConfig
from nemo.lightning.pytorch.callbacks.megatron_comm_overlap import MegatronCommOverlapCallback
from nemo.lightning.pytorch.optim import CosineAnnealingScheduler
from nemo.lightning.pytorch.optim.megatron import MegatronOptimizerModule
from nemo.utils.exp_manager import TimingCallback

NAME = "qwen25vl_32b"

HF_MODEL_NAME = "Qwen/Qwen2.5-VL-32B-Instruct"


@run.cli.factory(name=NAME)
def model() -> run.Config[pl.LightningModule]:
    """
    Factory function to create a Qwen2.5VL 32B model configuration.

    Returns:
        run.Config[pl.LightningModule]: Configuration for the Qwen2.5VL 32B model.

    Examples:
        CLI usage:
            $ nemo llm finetune model=qwen25vl_32b ...

        Python API usage:
            >>> model_config = model()
            >>> print(model_config)
    """

    return run.Config(vlm.Qwen2VLModel, config=run.Config(vlm.Qwen2VLConfig))


@run.cli.factory(target=llm.finetune, name=NAME)
def finetune_recipe(
    dir: Optional[str] = None,
    name: str = "default",
    num_nodes: int = 4,
    num_gpus_per_node: int = 8,
    peft_scheme: Optional[str] = 'none',
) -> run.Partial:
    """
    Create a finetuning recipe for Qwen2.5VL 32B model.

    This function sets up a complete configuration for finetuning, including
    model, trainer, data, logging, optimization, and resumption settings.

    Args:
        dir (Optional[str]): Directory for saving logs and checkpoints.
        name (str): Name of the finetuning run.
        num_nodes (int): Number of compute nodes to use.
        num_gpus_per_node (int): Number of GPUs per node.

    Returns:
        run.Partial: Partial configuration for finetuning.

    Examples:
        CLI usage:
            $ nemo llm finetune --factory qwen25vl_32b

        Python API usage:
            >>> recipe = finetune_recipe(name="qwen25vl_32b_finetune", num_nodes=1)
            >>> print(recipe)

    Note:
        This recipe uses the Mock dataset for finetuning.
    """

    # Training strategy setup
    strategy = run.Config(
        nl.MegatronStrategy,
        tensor_model_parallel_size=4,
        context_parallel_size=1,
        pipeline_model_parallel_size=2,
        pipeline_dtype=torch.bfloat16,
        sequence_parallel=True,
        ddp=DistributedDataParallelConfig(
            check_for_nan_in_grad=True,
            grad_reduce_in_fp32=True,
            overlap_grad_reduce=True,
            overlap_param_gather=True,
            average_in_collective=True,
        ),
        use_te_rng_tracker=True,
    )

    # tp_comm_overlap callback
    tp_comm_overlap_callback = run.Config(
        MegatronCommOverlapCallback, tp_comm_overlap=True, tp_comm_bootstrap_backend='nccl'
    )
    # Trainer setup
    trainer = run.Config(
        nl.Trainer,
        num_nodes=num_nodes,
        devices=num_gpus_per_node,
        max_steps=10,
        accelerator="gpu",
        strategy=strategy,
        plugins=run.Config(nl.MegatronMixedPrecision, precision="bf16-mixed"),
        callbacks=[run.Config(TimingCallback), tp_comm_overlap_callback],
        val_check_interval=10,
        limit_val_batches=0.0,
        log_every_n_steps=1,
        num_sanity_val_steps=0,
    )
    max_sequence_length = 8192

    data = run.Config(
        vlm.Qwen2VLMockDataModule,
        seq_length=max_sequence_length,
        global_batch_size=64,
        micro_batch_size=1,
        tokenizer=run.Config(AutoTokenizer, HF_MODEL_NAME),
        image_processor=run.Config(Qwen2VLImageProcessor),
        num_workers=1,
    )

    # Submodules configurations
    language_transformer_config = run.Config(llm.Qwen25Config32B, seq_length=max_sequence_length)
    vision_transformer_config = run.Config(Qwen25VLVisionConfig)
    vision_projection_config = run.Config(
        vlm.MultimodalProjectorConfig,
        projector_type="mcore_mlp",
        input_size=vision_transformer_config.hidden_size * (vision_transformer_config.spatial_merge_size**2),
        hidden_size=language_transformer_config.hidden_size,
        ffn_hidden_size=vision_transformer_config.hidden_size * (vision_transformer_config.spatial_merge_size**2),
    )

    # Qwen25VL model configuration
    qwen25vl_config = run.Config(
        vlm.Qwen2VLConfig,
        language_transformer_config=language_transformer_config,
        vision_transformer_config=vision_transformer_config,
        vision_projection_config=vision_projection_config,
        freeze_language_model=False,
        freeze_vision_model=True,
        enable_cuda_graph=True,
        use_te_rng_tracker=True,
        gradient_accumulation_fusion=True,
        cross_entropy_loss_fusion=True,
        bias_activation_fusion=True,
        bias_dropout_fusion=True,
        masked_softmax_fusion=True,
        attention_softmax_in_fp32=True,
        apply_rope_fusion=True,
        tp_comm_overlap=True,
        tp_comm_overlap_rs_dgrad=True,
        overlap_p2p_comm=True,
    )
    model = run.Config(
        vlm.Qwen2VLModel,
        config=qwen25vl_config,
        model_version="qwen25-vl",
        tokenizer=run.Config(AutoTokenizer, HF_MODEL_NAME),
    )

    nemo_resume = run.Config(
        nl.AutoResume,
        resume_if_exists=True,
        resume_ignore_no_checkpoint=True,
    )
    # Optimizer and scheduler setup
    opt_config = run.Config(
        OptimizerConfig,
        optimizer='adam',
        lr=2.0e-06,
        adam_beta1=0.9,
        adam_beta2=0.95,
        use_distributed_optimizer=True,
        bf16=True,
    )
    sched = run.Config(
        CosineAnnealingScheduler,
        max_steps=10,
        warmup_steps=0,
        constant_steps=1000,
        min_lr=1.0e-07,
    )
    opt = run.Config(MegatronOptimizerModule, opt_config, sched)

    recipe = run.Partial(
        llm.finetune,
        model=model,
        trainer=trainer,
        data=data,
        log=llm.default_log(dir=dir, name=name, tensorboard_logger=tensorboard_logger(name=name)),
        optim=opt,
        resume=nemo_resume,
    )
    return recipe


if __name__ == "__main__":
    env_vars = {
        "CUDA_VISIBLE_DEVICES": "0,1,2,3,4,5,6,7",
    }
    recipe = finetune_recipe(num_gpus_per_node=8)

    executor = run.LocalExecutor(ntasks_per_node=8, launcher="torchrun", env_vars=env_vars)

    run.run(recipe, executor=executor, name="qwen2.5vl_32b_finetune")



# Copyright (c) 2023, NVIDIA CORPORATION & AFFILIATES.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import random
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, List, Optional

import librosa
import numpy as np
import torch.utils.data

from nemo.collections.asr.parts.utils.manifest_utils import read_manifest
from nemo.collections.common.tokenizers.text_to_speech.tts_tokenizers import BaseTokenizer
from nemo.collections.tts.parts.preprocessing.feature_processors import FeatureProcessor
from nemo.collections.tts.parts.preprocessing.features import Featurizer
from nemo.collections.tts.parts.utils.tts_dataset_utils import (
    _read_audio,
    beta_binomial_prior_distribution,
    chunk_and_tokenize_text_by_sentence,
    filter_dataset_by_duration,
    get_weighted_sampler,
    load_audio,
    stack_tensors,
)
from nemo.core.classes import Dataset
from nemo.utils import logging


@dataclass
class DatasetMeta:
    manifest_path: Path
    audio_dir: Path
    feature_dir: Path = None
    sample_weight: float = 1.0
    tokenizer_names: List[str] = None


@dataclass
class DatasetSample:
    dataset_name: str
    manifest_entry: Dict[str, Any]
    audio_dir: Optional[Path]
    feature_dir: Optional[Path]
    text: str
    speaker: str
    speaker_index: int = None
    tokenizer_names: List[str] = None


class TextToSpeechDataset(Dataset):
    """
    Class for processing and loading text to speech training examples.

    Args:
        dataset_meta: Dict of dataset names (string) to dataset metadata.
        sample_rate: Sample rate to load audio as. If the audio is stored at a different sample rate, then it will
            be resampled.
        text_tokenizer: Tokenizer to apply to the text field.
        weighted_sampling_steps_per_epoch: Optional int, If provided, then data will be sampled (with replacement) based on
            the sample weights provided in the dataset metadata. If None, then sample weights will be ignored.
        speaker_path: Optional, path to JSON file with speaker indices, for multi-speaker training. Can be created with
            scripts.dataset_processing.tts.create_speaker_map.py
        featurizers: Optional, list of featurizers to load feature data from. Should be the same config provided
            when running scripts.dataset_processing.tts.compute_features.py before training.
        feature_processors: Optional, list of feature processors to run on training examples.
        align_prior_hop_length: Optional int, hop length of audio features.
            If provided alignment prior will be calculated and included in batch output. Must match hop length
            of audio features used for training.
        min_duration: Optional float, if provided audio files in the training manifest shorter than 'min_duration'
            will be ignored.
        max_duration: Optional float, if provided audio files in the training manifest longer than 'max_duration'
            will be ignored.
        volume_norm: Whether to apply volume normalization to loaded audio.
    """

    def __init__(
        self,
        dataset_meta: Dict,
        sample_rate: int,
        text_tokenizer: BaseTokenizer,
        weighted_sampling_steps_per_epoch: Optional[int] = None,
        speaker_path: Optional[Path] = None,
        featurizers: Optional[Dict[str, Featurizer]] = None,
        feature_processors: Optional[Dict[str, FeatureProcessor]] = None,
        align_prior_hop_length: Optional[int] = None,
        min_duration: Optional[float] = None,
        max_duration: Optional[float] = None,
        volume_norm: bool = True,
    ):
        super().__init__()

        self.sample_rate = sample_rate
        self.text_tokenizer = text_tokenizer
        self.weighted_sampling_steps_per_epoch = weighted_sampling_steps_per_epoch
        self.align_prior_hop_length = align_prior_hop_length
        self.include_align_prior = self.align_prior_hop_length is not None
        self.volume_norm = volume_norm

        if speaker_path:
            self.include_speaker = True
            with open(speaker_path, 'r', encoding="utf-8") as speaker_f:
                speaker_index_map = json.load(speaker_f)
        else:
            self.include_speaker = False
            speaker_index_map = None

        if featurizers:
            logging.info(f"Found featurizers {featurizers.keys()}")
            self.featurizers = list(featurizers.values())
        else:
            self.featurizers = []

        if feature_processors:
            logging.info(f"Found featurize processors {feature_processors.keys()}")
            self.feature_processors = list(feature_processors.values())
        else:
            self.feature_processors = []

        self.data_samples = []
        self.sample_weights = []
        for dataset_name, dataset_info in dataset_meta.items():
            dataset = DatasetMeta(**dataset_info)
            samples, weights = self._preprocess_manifest(
                dataset_name=dataset_name,
                dataset=dataset,
                min_duration=min_duration,
                max_duration=max_duration,
                speaker_index_map=speaker_index_map,
            )
            self.data_samples += samples
            self.sample_weights += weights

    def get_sampler(self, batch_size: int, world_size: int) -> Optional[torch.utils.data.Sampler]:
        if not self.weighted_sampling_steps_per_epoch:
            return None

        sampler = get_weighted_sampler(
            sample_weights=self.sample_weights,
            batch_size=batch_size,
            world_size=world_size,
            num_steps=self.weighted_sampling_steps_per_epoch,
        )
        return sampler

    def _preprocess_manifest(
        self,
        dataset_name: str,
        dataset: DatasetMeta,
        min_duration: float,
        max_duration: float,
        speaker_index_map: Dict[str, int],
    ):
        entries = read_manifest(dataset.manifest_path)
        filtered_entries, total_hours, filtered_hours = filter_dataset_by_duration(
            entries=entries, min_duration=min_duration, max_duration=max_duration
        )

        logging.info(dataset_name)
        logging.info(f"Original # of files: {len(entries)}")
        logging.info(f"Filtered # of files: {len(filtered_entries)}")
        logging.info(f"Original duration: {total_hours:.2f} hours")
        logging.info(f"Filtered duration: {filtered_hours:.2f} hours")

        samples = []
        sample_weights = []
        for entry in filtered_entries:

            if "normalized_text" in entry:
                text = entry["normalized_text"]
            else:
                text = entry["text"]

            if self.include_speaker:
                speaker = entry["speaker"]
                speaker_index = speaker_index_map[speaker]
            else:
                speaker = None
                speaker_index = 0

            sample = DatasetSample(
                dataset_name=dataset_name,
                manifest_entry=entry,
                audio_dir=None if dataset.audio_dir is None else Path(dataset.audio_dir),
                feature_dir=None if dataset.feature_dir is None else Path(dataset.feature_dir),
                text=text,
                speaker=speaker,
                speaker_index=speaker_index,
                tokenizer_names=dataset.tokenizer_names,
            )
            samples.append(sample)
            sample_weights.append(dataset.sample_weight)

        return samples, sample_weights

    def __len__(self):
        return len(self.data_samples)

    def __getitem__(self, index):
        data = self.data_samples[index]

        audio_array, _, audio_filepath_rel = load_audio(
            manifest_entry=data.manifest_entry,
            audio_dir=data.audio_dir,
            sample_rate=self.sample_rate,
            volume_norm=self.volume_norm,
        )
        audio = torch.tensor(audio_array, dtype=torch.float32)
        audio_len = audio.shape[0]

        tokens = self.text_tokenizer(data.text)
        tokens = torch.tensor(tokens, dtype=torch.int32)
        text_len = tokens.shape[0]

        example = {
            "dataset_name": data.dataset_name,
            "audio_filepath": audio_filepath_rel,
            "audio": audio,
            "audio_len": audio_len,
            "tokens": tokens,
            "text_len": text_len,
        }

        if data.speaker is not None:
            example["speaker"] = data.speaker
            example["speaker_index"] = data.speaker_index

        if self.include_align_prior:
            spec_len = 1 + librosa.core.samples_to_frames(audio_len, hop_length=self.align_prior_hop_length)
            align_prior = beta_binomial_prior_distribution(phoneme_count=text_len, mel_count=spec_len)
            align_prior = torch.tensor(align_prior, dtype=torch.float32)
            example["align_prior"] = align_prior

        for featurizer in self.featurizers:
            feature_dict = featurizer.load(
                manifest_entry=data.manifest_entry, audio_dir=data.audio_dir, feature_dir=data.feature_dir
            )
            example.update(feature_dict)

        for processor in self.feature_processors:
            processor.process(example)

        return example

    def collate_fn(self, batch: List[dict]):
        dataset_name_list = []
        audio_filepath_list = []
        audio_list = []
        audio_len_list = []
        token_list = []
        token_len_list = []
        speaker_list = []
        prior_list = []

        for example in batch:
            dataset_name_list.append(example["dataset_name"])
            audio_filepath_list.append(example["audio_filepath"])

            audio_list.append(example["audio"])
            audio_len_list.append(example["audio_len"])

            token_list.append(example["tokens"])
            token_len_list.append(example["text_len"])

            if self.include_speaker:
                speaker_list.append(example["speaker_index"])

            if self.include_align_prior:
                prior_list.append(example["align_prior"])

        batch_audio_len = torch.IntTensor(audio_len_list)
        audio_max_len = int(batch_audio_len.max().item())

        batch_token_len = torch.IntTensor(token_len_list)
        token_max_len = int(batch_token_len.max().item())

        batch_audio = stack_tensors(audio_list, max_lens=[audio_max_len])
        batch_tokens = stack_tensors(token_list, max_lens=[token_max_len], pad_value=self.text_tokenizer.pad)

        batch_dict = {
            "dataset_names": dataset_name_list,
            "audio_filepaths": audio_filepath_list,
            "audio": batch_audio,
            "audio_lens": batch_audio_len,
            "text": batch_tokens,
            "text_lens": batch_token_len,
        }

        if self.include_speaker:
            batch_dict["speaker_id"] = torch.IntTensor(speaker_list)

        if self.include_align_prior:
            spec_max_len = max([prior.shape[0] for prior in prior_list])
            text_max_len = max([prior.shape[1] for prior in prior_list])
            batch_dict["align_prior_matrix"] = stack_tensors(
                prior_list,
                max_lens=[text_max_len, spec_max_len],
            )

        for featurizer in self.featurizers:
            feature_dict = featurizer.collate_fn(batch)
            batch_dict.update(feature_dict)

        return batch_dict


class MagpieTTSDataset(TextToSpeechDataset):
    """
    Class for processing and loading text to speech training examples for Magpie-TTS model.
    In addition to the manifest structure for TextToSpeechDataset, we can have the following keys:
    context_audio_filepath, context_audio_duration, target_audio_codes_path, context_audio_codes_path.
    Note: target_audio_codes_path, context_audio_codes_path are absolute paths to the cached audio codes.
    If they are not present in the manifest or if load_cached_codes_if_available=False, then the audio will be loaded
    and codes will be computed on the fly in the model class.

    Args:
        dataset_meta: Dict of dataset names (string) to dataset metadata.
        sample_rate: Sample rate to load audio as. If the audio is stored at a different sample rate, then it will
            be resampled.
        weighted_sampling_steps_per_epoch: Optional int, If provided, then data will be sampled (with replacement) based on
            the sample weights provided in the dataset metadata. If None, then sample weights will be ignored.
        min_duration: Optional float, if provided audio files in the training manifest shorter than 'min_duration'
            will be ignored.
        max_duration: Optional float, if provided audio files in the training manifest longer than 'max_duration'
            will be ignored.
        volume_norm: Whether to apply volume normalization to loaded audio.
        codec_model_samples_per_frame: Num samples in waveform per codec frame (codec downsample factor).
        bos_id: Text BOS token id.
        eos_id: Text EOS token id.
        num_audio_codebooks: Number of audio codebooks.
        prior_scaling_factor: Scaling factor for the beta binomial prior distribution.
        load_cached_codes_if_available: Whether to load cached audio codes if available *_codes_path keys are available in the manifest.
        dataset_type: Dataset type (train, dev, test).
        tokenizer_config: Config of the tokenzizer used in worker_init_fn to setup the tokenizer (See Magpie-TTS yamls)
        load_16khz_audio: Whether to load 16khz audio for SV model.
        use_text_conditioning_tokenizer: Set True for text context conditioning.
        pad_context_text_to_max_duration: Whether to pad context text to max context audio frames.
        context_duration_min: Minimum duration of context audio in seconds.
        context_duration_max: Maximum duration of context audio in seconds.
        text_context_remapping: Dict defining mapping of multiple text contexts to a single text context.
        text_context_remapping_prob: Probability of remapping the original text context to a remapped text context.
    """

    def __init__(
        self,
        dataset_meta: Dict,
        sample_rate: int,
        weighted_sampling_steps_per_epoch: Optional[int] = None,
        min_duration: Optional[float] = None,
        max_duration: Optional[float] = None,
        volume_norm: bool = True,
        codec_model_samples_per_frame: int = None,
        bos_id: int = None,
        eos_id: int = None,
        num_audio_codebooks: int = None,
        prior_scaling_factor: float = None,
        load_cached_codes_if_available: bool = True,
        dataset_type: str = 'train',
        tokenizer_config=None,
        load_16khz_audio: bool = True,
        use_text_conditioning_tokenizer: bool = False,
        text_conditioning_tokenizer_name: str = None,
        pad_context_text_to_max_duration: bool = False,
        context_duration_min: float = 3.0,
        context_duration_max: float = 10.0,
        text_context_remapping: Dict[str, str] = None,
        text_context_remapping_prob: float = 0.0,
    ):
        super().__init__(
            dataset_meta=dataset_meta,
            sample_rate=sample_rate,
            text_tokenizer=None,
            weighted_sampling_steps_per_epoch=weighted_sampling_steps_per_epoch,
            speaker_path=None,
            featurizers=None,
            feature_processors=None,
            align_prior_hop_length=None,
            min_duration=min_duration,
            max_duration=max_duration,
            volume_norm=volume_norm,
        )
        self.bos_id = bos_id  # TODO @xueyang: this should be removed since no other places used it.
        self.eos_id = eos_id
        self.num_audio_codebooks = num_audio_codebooks
        self.codec_model_samples_per_frame = codec_model_samples_per_frame
        self.include_align_prior = prior_scaling_factor is not None
        self.prior_scaling_factor = prior_scaling_factor
        self.load_cached_codes_if_available = load_cached_codes_if_available
        self.dataset_type = dataset_type
        self.tokenizer_config = tokenizer_config
        self.text_tokenizer = None  # Assigned in worker_init_fn in model file
        self.load_16khz_audio = load_16khz_audio
        self.use_text_conditioning_tokenizer = use_text_conditioning_tokenizer
        self.text_conditioning_tokenizer_name = text_conditioning_tokenizer_name
        self.pad_context_text_to_max_duration = pad_context_text_to_max_duration
        self.context_duration_min = context_duration_min
        self.context_duration_max = context_duration_max
        self.text_context_remapping = text_context_remapping
        self.text_context_remapping_prob = text_context_remapping_prob

    def get_num_audio_samples_to_slice(self, duration, sample_rate):
        num_codec_frames = int(duration * sample_rate / self.codec_model_samples_per_frame)
        num_audio_samples = num_codec_frames * self.codec_model_samples_per_frame
        return num_audio_samples

    def __getitem__(self, index):
        data = self.data_samples[index]
        tokenizer_name = "english_phoneme"  # Default to english phoneme tokenizer
        if data.tokenizer_names is not None:
            # Pick a random tokenizer from the list of tokenizers
            tokenizer_name = random.choice(data.tokenizer_names)
        tokens = self.text_tokenizer.encode(text=data.text, tokenizer_name=tokenizer_name)
        tokens = tokens + [self.eos_id]  # Not adding BOS id
        tokens = torch.tensor(tokens, dtype=torch.int32)
        text_len = tokens.shape[0]

        example = {
            "dataset_name": data.dataset_name,
            "tokens": tokens,
            "text_len": text_len,
        }

        if self.load_cached_codes_if_available and 'target_audio_codes_path' in data.manifest_entry:
            audio_codes_path = data.manifest_entry['target_audio_codes_path']
            audio_codes = torch.load(audio_codes_path)  # (C, T)
            spec_len = audio_codes.shape[1] + 1  # +1 for EOS
            audio_codes_len = audio_codes.shape[1]
            example['audio_codes'] = audio_codes
            example['audio_codes_len'] = audio_codes_len
            example['audio_filepath'] = audio_codes_path
            if 'audio_filepath' in data.manifest_entry:
                # If audio_filepath is available, then use the actual audio file path.
                example['audio_filepath'] = data.manifest_entry['audio_filepath']
        elif 'audio_filepath' in data.manifest_entry:
            # Only load audio if codes are not available
            audio_array, _, audio_filepath_rel = load_audio(
                manifest_entry=data.manifest_entry,
                audio_dir=data.audio_dir,
                sample_rate=self.sample_rate,
                volume_norm=self.volume_norm,
            )
            audio = torch.tensor(audio_array, dtype=torch.float32)
            # Pad audio to be multiple of downsample factor
            audio = torch.nn.functional.pad(
                audio,
                (0, self.codec_model_samples_per_frame - (audio.shape[0] % self.codec_model_samples_per_frame)),
                value=0,
            )
            audio_len = audio.shape[0]
            example['audio_filepath'] = data.manifest_entry['audio_filepath']
            example['audio'] = audio
            example['audio_len'] = audio_len
            spec_len = int(audio_len / self.codec_model_samples_per_frame) + 1  # +1 for EOS

        if self.load_cached_codes_if_available and 'context_audio_codes_path' in data.manifest_entry:
            context_audio_codes_path = data.manifest_entry['context_audio_codes_path']
            context_audio_codes = torch.load(context_audio_codes_path)  # (8, T)
            # Sample random duration between self.context_duration_min and self.context_duration_max
            _context_duration_to_slice = random.uniform(self.context_duration_min, self.context_duration_max)
            _num_frames_to_slice = int(
                _context_duration_to_slice * self.sample_rate / self.codec_model_samples_per_frame
            )
            if _num_frames_to_slice < context_audio_codes.shape[1]:
                start_idx = random.randint(0, context_audio_codes.shape[1] - _num_frames_to_slice)
                context_audio_codes = context_audio_codes[:, start_idx : start_idx + _num_frames_to_slice]
            else:
                # Repeaet the audio if it is shorter than the desired duration
                _num_repeats = int(np.ceil(_num_frames_to_slice / context_audio_codes.shape[1]))
                # context_audio_codes is a tensor of shape (num_codebooks, T)
                context_audio_codes_repeated = context_audio_codes.repeat(1, _num_repeats)
                context_audio_codes = context_audio_codes_repeated[:, :_num_frames_to_slice]

            context_audio_codes_len = context_audio_codes.shape[1]
            example['context_audio_codes'] = context_audio_codes
            example['context_audio_codes_len'] = context_audio_codes_len
        elif 'context_audio_filepath' in data.manifest_entry:
            context_audio_filepath = os.path.join(data.audio_dir, data.manifest_entry['context_audio_filepath'])
            context_duration = data.manifest_entry['context_audio_duration']
            context_audio_array = _read_audio(
                audio_filepath=context_audio_filepath,
                sample_rate=self.sample_rate,
                offset=0,
                duration=context_duration,
            )
            context_audio_array = context_audio_array.samples
            _context_duration_to_slice = random.uniform(self.context_duration_min, self.context_duration_max)
            _num_samples_to_slice = self.get_num_audio_samples_to_slice(_context_duration_to_slice, self.sample_rate)
            if _num_samples_to_slice < len(context_audio_array):
                start_idx = random.randint(0, len(context_audio_array) - _num_samples_to_slice)
                context_audio_array = context_audio_array[start_idx : start_idx + _num_samples_to_slice]
            else:
                # Repeaet the audio if it is shorter than the desired duration
                _num_repeats = int(np.ceil(_num_samples_to_slice / len(context_audio_array)))
                context_audio_array = np.tile(context_audio_array, _num_repeats)
                context_audio_array = context_audio_array[:_num_samples_to_slice]
            context_audio = torch.tensor(context_audio_array, dtype=torch.float32)
            context_audio_len = context_audio.shape[0]
            example['context_audio'] = context_audio
            example['context_audio_len'] = context_audio_len
        else:
            # We always want to have context_audio_codes if available for multi-encoder model. These are ignored for singlencoder model.
            # If context audio is not available, just use a dummy context_audio_codes
            # (Will be used in text context scenario)
            if self.load_cached_codes_if_available:
                context_audio_codes = torch.zeros([self.num_audio_codebooks, 0], dtype=torch.int32)
                context_audio_codes_len = 0
                example['context_audio_codes'] = context_audio_codes
                example['context_audio_codes_len'] = context_audio_codes_len
            else:
                # @shehzeenh: Added this condition so that a batch does not have a mix of context_audio and context_audio_codes
                # @blisc: Added a +1. If we send in exactly 882 samples, then a conv layer complains about padding.
                #         Adding 883 works. This occurs when we use text context during inference.
                context_audio = torch.zeros(self.codec_model_samples_per_frame + 1, dtype=torch.float32)
                context_audio_len = context_audio.shape[0]
                example['context_audio'] = context_audio
                example['context_audio_len'] = context_audio_len

        # 16kHz audio is used for SV model
        if self.load_16khz_audio:
            if 'context_audio_filepath' in data.manifest_entry:
                # If context_audio_filepath is available, then use that for 16khz audio for SV model
                context_audio_filepath = os.path.join(data.audio_dir, data.manifest_entry['context_audio_filepath'])
                context_duration = data.manifest_entry['context_audio_duration']
                audio_array_16khz = _read_audio(
                    audio_filepath=context_audio_filepath, sample_rate=16000, offset=0, duration=context_duration
                ).samples
            else:
                # Otherwise, load the target audio file.
                audio_array_16khz, _, _ = load_audio(
                    manifest_entry=data.manifest_entry,
                    audio_dir=data.audio_dir,
                    sample_rate=16000,
                    volume_norm=self.volume_norm,
                )
            _context_duration_to_slice = random.uniform(self.context_duration_min, self.context_duration_max)
            _num_samples_to_slice = int(_context_duration_to_slice * 16000)
            if _num_samples_to_slice < len(audio_array_16khz):
                start_idx = random.randint(0, len(audio_array_16khz) - _num_samples_to_slice)
                audio_array_16khz = audio_array_16khz[start_idx : start_idx + _num_samples_to_slice]
            audio_16khz = torch.tensor(audio_array_16khz, dtype=torch.float32)
            audio_len_16khz = audio_16khz.shape[0]
            example['audio_16khz'] = audio_16khz
            example['audio_len_16khz'] = audio_len_16khz

        if self.use_text_conditioning_tokenizer:
            if 'context_text' in data.manifest_entry:
                context_text = data.manifest_entry['context_text']
                if self.text_context_remapping is not None and context_text in self.text_context_remapping:
                    if self.dataset_type == 'train' and random.random() < self.text_context_remapping_prob:
                        # Only remap during training. Give the exact text context during inference.
                        context_text = self.text_context_remapping[context_text]
                context_tokens = self.text_tokenizer.encode(context_text, self.text_conditioning_tokenizer_name)
                example['has_text_context'] = True
            else:
                context_tokens = self.text_tokenizer.encode("[NO TEXT CONTEXT]", self.text_conditioning_tokenizer_name)
                example['has_text_context'] = False
            if self.pad_context_text_to_max_duration:
                _required_len = (
                    int(self.context_duration_max * self.sample_rate / self.codec_model_samples_per_frame) + 2
                )  # +2 for BOS and EOS
                if len(context_tokens) < _required_len:
                    _pad_id = self.text_tokenizer.tokenizer_pad_ids[self.text_conditioning_tokenizer_name]
                    context_tokens += [_pad_id] * (_required_len - len(context_tokens))
                else:
                    context_tokens = context_tokens[:_required_len]

            context_tokens = torch.tensor(context_tokens, dtype=torch.int32)
            context_text_len = context_tokens.shape[0]
            example['context_text_tokens'] = context_tokens
            example['context_text_len'] = context_text_len

        if self.include_align_prior:
            align_prior = beta_binomial_prior_distribution(
                phoneme_count=text_len, mel_count=spec_len, scaling_factor=self.prior_scaling_factor
            )
            align_prior = torch.tensor(align_prior, dtype=torch.float32)
            example["align_prior"] = align_prior

        if "original_text" in data.manifest_entry:
            # Raw Text is used as the GT for CER/WER computation in DPO pref data generation
            # and GRPO reward setup. For manifests in which the 'text' field is phonemized,
            # we use the 'original_text' field as the raw text. Otherwise, we use the regular text field.
            example['raw_text'] = data.manifest_entry['original_text']
        else:
            example['raw_text'] = data.text

        example['language'] = data.manifest_entry.get('language', 'en')

        if "reward" in data.manifest_entry:
            example["reward"] = data.manifest_entry["reward"]

        # Get speaker index if available (for models with baked context embeddings)
        if 'speaker_index' in data.manifest_entry:
            example['speaker_index'] = data.manifest_entry['speaker_index']

        return example

    def collate_fn(self, batch: List[dict]):
        dataset_name_list = []
        audio_filepath_list = []
        audio_list = []
        audio_len_list = []
        audio_list_16khz = []
        audio_len_list_16khz = []
        token_list = []
        token_len_list = []
        prior_list = []
        audio_codes_list = []
        audio_codes_len_list = []
        context_audio_list = []
        context_audio_len_list = []
        context_audio_codes_list = []
        context_audio_codes_len_list = []
        context_text_tokens_list = []
        context_text_tokens_len_list = []
        context_has_text_context_list = []
        reward_list = []
        raw_text_list = []
        language_list = []
        speaker_indices_list = []
        for example in batch:
            dataset_name_list.append(example["dataset_name"])
            raw_text_list.append(example["raw_text"])
            language_list.append(example["language"])

            token_list.append(example["tokens"])
            token_len_list.append(example["text_len"])

            if 'audio_filepath' in example:
                audio_filepath_list.append(example["audio_filepath"])

            if 'audio' in example:
                audio_list.append(example["audio"])
                audio_len_list.append(example["audio_len"])

            if 'audio_16khz' in example:
                audio_list_16khz.append(example["audio_16khz"])
                audio_len_list_16khz.append(example["audio_len_16khz"])

            if 'audio_codes' in example:
                audio_codes_list.append(example['audio_codes'])
                audio_codes_len_list.append(example['audio_codes_len'])

            if 'context_audio' in example:
                context_audio_list.append(example['context_audio'])
                context_audio_len_list.append(example['context_audio_len'])

            if 'context_audio_codes' in example:
                context_audio_codes_list.append(example['context_audio_codes'])
                context_audio_codes_len_list.append(example['context_audio_codes_len'])

            if 'context_text_tokens' in example:
                context_text_tokens_list.append(example['context_text_tokens'])
                context_text_tokens_len_list.append(example['context_text_len'])
                context_has_text_context_list.append(example['has_text_context'])

            if 'reward' in example:
                reward_list.append(example['reward'])

            if 'speaker_index' in example:
                speaker_indices_list.append(example['speaker_index'])

            if self.include_align_prior:
                prior_list.append(example["align_prior"])

        batch_token_len = torch.IntTensor(token_len_list)
        token_max_len = int(batch_token_len.max().item())
        batch_tokens = stack_tensors(token_list, max_lens=[token_max_len], pad_value=self.text_tokenizer.pad)

        batch_dict = {
            "dataset_names": dataset_name_list,
            "raw_texts": raw_text_list,
            "languages": language_list,
            "audio_filepaths": audio_filepath_list,
            "text": batch_tokens,
            "text_lens": batch_token_len,
        }

        if len(audio_list) > 0:
            batch_audio_len = torch.IntTensor(audio_len_list)
            audio_max_len = int(batch_audio_len.max().item())
            batch_audio = stack_tensors(audio_list, max_lens=[audio_max_len])
            batch_dict['audio'] = batch_audio
            batch_dict['audio_lens'] = batch_audio_len

        if len(audio_list_16khz) > 0:
            batch_audio_len_16khz = torch.IntTensor(audio_len_list_16khz)
            audio_max_len_16khz = int(batch_audio_len_16khz.max().item())
            batch_audio_16khz = stack_tensors(audio_list_16khz, max_lens=[audio_max_len_16khz])
            batch_dict['audio_16khz'] = batch_audio_16khz
            batch_dict['audio_lens_16khz'] = batch_audio_len_16khz

        if len(audio_codes_list) > 0:
            batch_audio_codes_len = torch.IntTensor(audio_codes_len_list)
            audio_codes_max_len = int(batch_audio_codes_len.max().item())
            batch_audio_codes = stack_tensors(audio_codes_list, max_lens=[audio_codes_max_len])
            batch_dict['audio_codes'] = batch_audio_codes
            batch_dict['audio_codes_lens'] = batch_audio_codes_len

        if len(context_audio_list) > 0:
            batch_context_audio_len = torch.IntTensor(context_audio_len_list)
            context_audio_max_len = int(batch_context_audio_len.max().item())
            batch_context_audio = stack_tensors(context_audio_list, max_lens=[context_audio_max_len])
            batch_dict['context_audio'] = batch_context_audio
            batch_dict['context_audio_lens'] = batch_context_audio_len

        if len(context_audio_codes_list) > 0:
            batch_context_audio_codes_len = torch.IntTensor(context_audio_codes_len_list)
            context_audio_codes_max_len = int(batch_context_audio_codes_len.max().item())
            # TODO @xueyang: verify if batch_context_audio_codes are integer.
            batch_context_audio_codes = stack_tensors(context_audio_codes_list, max_lens=[context_audio_codes_max_len])
            batch_dict['context_audio_codes'] = batch_context_audio_codes
            batch_dict['context_audio_codes_lens'] = batch_context_audio_codes_len

        if self.use_text_conditioning_tokenizer:
            batch_context_text_tokens_len = torch.IntTensor(context_text_tokens_len_list)
            context_text_tokens_max_len = int(batch_context_text_tokens_len.max().item())
            # TODO @xueyang: potential bugs if self.tokenizer.pad is not 0.0. verify if batch_context_text_tokens are integer.
            batch_context_text_tokens = stack_tensors(context_text_tokens_list, max_lens=[context_text_tokens_max_len])
            batch_dict['context_text_tokens'] = batch_context_text_tokens
            batch_dict['context_text_tokens_lens'] = batch_context_text_tokens_len
            batch_dict['has_text_context'] = torch.BoolTensor(context_has_text_context_list)

        if self.include_align_prior:
            spec_max_len = max([prior.shape[0] for prior in prior_list])
            text_max_len = max([prior.shape[1] for prior in prior_list])
            batch_dict["align_prior_matrix"] = stack_tensors(
                prior_list,
                max_lens=[text_max_len, spec_max_len],
            )

        if len(reward_list) > 0:
            batch_dict['rewards'] = torch.FloatTensor(reward_list)

        if len(speaker_indices_list) > 0:
            batch_dict['speaker_indices'] = torch.tensor(speaker_indices_list, dtype=torch.int64)

        # Assert no more than one of audio or audio_codes in the batch
        if 'audio' in batch_dict:
            assert 'audio_codes' not in batch_dict

        # Assert no more than one of context_audio or context_audio_codes in the batch
        if 'context_audio' in batch_dict:
            assert 'context_audio_codes' not in batch_dict

        return batch_dict


class MagpieTTSDatasetDPO(MagpieTTSDataset):
    """
    This class is meant to be used with the DPO model. To generate manifests for this dataset, please use
        - scripts/magpietts/dpo/create_text_contextpairs.py
        - scripts/magpietts/dpo/create_preference_pairs.py
    in sequence to generate samples and create preference pairs.
    """

    def __len__(self):
        return len(self.data_samples) // 2

    def __getitem__(self, index):
        chosen_example = super().__getitem__(index * 2)
        rejected_example = super().__getitem__(index * 2 + 1)
        assert chosen_example['reward'] == 1.0
        assert rejected_example['reward'] < 1.0
        return {"chosen": chosen_example, "rejected": rejected_example}

    def collate_fn(self, batch: List[dict]):
        chosen_batch = [example['chosen'] for example in batch]
        rejected_batch = [example['rejected'] for example in batch]
        chosen_collated = super().collate_fn(chosen_batch)
        rejected_collated = super().collate_fn(rejected_batch)
        return {"chosen": chosen_collated, "rejected": rejected_collated}


class LongFormTTSInferenceDataset(MagpieTTSDataset):
    """
    Dataset for longform TTS inference with sentence-level text chunking.

    Inherits from MagpieTTSDataset to reuse context audio loading, text conditioning,
    and other preprocessing logic. Adds sentence-level text chunking on top.

    Args:
        dataset_meta: Dataset metadata dictionary (same format as MagpieTTSDataset).
        sample_rate: Audio sample rate.
        tokenizer_name: Name of the tokenizer to use for sentence chunking.
        codec_model_samples_per_frame: Samples per codec frame.
        eos_id: End-of-sequence token ID.
        audio_bos_id: Audio BOS token ID (for target audio).
        audio_eos_id: Audio EOS token ID (for target audio).
        context_audio_bos_id: Context audio BOS token ID.
        context_audio_eos_id: Context audio EOS token ID.
        num_audio_codebooks: Number of audio codebooks.
        context_duration_min: Minimum context duration in seconds.
        context_duration_max: Maximum context duration in seconds.
        use_text_conditioning_tokenizer: Whether model uses text conditioning encoder.
        text_conditioning_tokenizer_name: Name of text conditioning tokenizer.
        pad_context_text_to_max_duration: Whether to pad context text.
        load_16khz_audio: Whether to load 16kHz audio for SV model.
    """

    def __init__(
        self,
        dataset_meta: Dict[str, Any],
        sample_rate: int,
        tokenizer_name: str,
        codec_model_samples_per_frame: int,
        eos_id: int,
        num_audio_codebooks: int,
        context_duration_min: float = 3.0,
        context_duration_max: float = 10.0,
        use_text_conditioning_tokenizer: bool = False,
        text_conditioning_tokenizer_name: str = None,
        pad_context_text_to_max_duration: bool = False,
        load_16khz_audio: bool = False,
        **kwargs,
    ):
        # Initialize parent - handles manifest reading and context audio loading
        super().__init__(
            dataset_meta=dataset_meta,
            sample_rate=sample_rate,
            codec_model_samples_per_frame=codec_model_samples_per_frame,
            eos_id=eos_id,
            num_audio_codebooks=num_audio_codebooks,
            context_duration_min=context_duration_min,
            context_duration_max=context_duration_max,
            use_text_conditioning_tokenizer=use_text_conditioning_tokenizer,
            text_conditioning_tokenizer_name=text_conditioning_tokenizer_name,
            pad_context_text_to_max_duration=pad_context_text_to_max_duration,
            load_16khz_audio=load_16khz_audio,
            load_cached_codes_if_available=True,  # Prefer codes for inference
            dataset_type='test',
            **kwargs,
        )
        self.tokenizer_name = tokenizer_name

    def __getitem__(self, idx: int) -> Dict[str, Any]:
        """
        Add sentence chunking on top of parent's __getitem__.

        Returns:
            Dictionary containing all parent fields plus:
                - idx: Sample index
                - chunked_tokens: List of tokenized text chunks (per sentence)
                - chunked_tokens_len: List of token lengths
                - entry: Original manifest entry
        """
        # Get text for sentence chunking
        data = self.data_samples[idx]
        text = data.text  # entry.get("normalized_text", entry.get("text", ""))

        # Call parent to get ALL the context audio, text conditioning, etc.
        example = super().__getitem__(idx)

        # Sentence chunking (longform-specific)
        chunked_tokens, chunked_tokens_len, _ = chunk_and_tokenize_text_by_sentence(
            text,
            self.tokenizer_name,
            self.text_tokenizer,
            self.eos_id,
            language=example['language'],
        )

        # Handle empty text edge case
        if not chunked_tokens:
            chunked_tokens = [torch.tensor([self.eos_id], dtype=torch.int32)]
            chunked_tokens_len = [1]

        # Add longform-specific fields
        example['idx'] = idx
        example['chunked_tokens'] = chunked_tokens
        example['chunked_tokens_len'] = chunked_tokens_len

        return example

    def collate_fn(self, batch: List[Dict[str, Any]]) -> Dict[str, Any]:
        """
        Collate function for batching longform samples.

        Calls parent's collate_fn to handle context audio, text conditioning, etc.,
        then adds longform-specific fields (chunked_tokens).
        """
        # Call parent's collate_fn to handle all standard fields
        batch_dict = super().collate_fn(batch)

        # Add longform-specific fields
        indices = []
        chunked_tokens_list = []
        chunked_tokens_lens_list = []

        # Find max number of chunks across batch
        max_num_chunks = max(len(sample['chunked_tokens']) for sample in batch)

        for sample in batch:
            indices.append(sample['idx'])

            # Pad chunked tokens to max_num_chunks with single EOS token
            num_padding = max_num_chunks - len(sample['chunked_tokens'])
            padded_tokens = sample['chunked_tokens'] + [
                torch.tensor([self.eos_id], dtype=torch.int32) for _ in range(num_padding)
            ]
            padded_lens = sample['chunked_tokens_len'] + [1] * num_padding

            chunked_tokens_list.append(padded_tokens)
            chunked_tokens_lens_list.append(padded_lens)

        # Add longform-specific fields to batch_dict
        batch_dict['idx'] = indices
        batch_dict['chunked_tokens'] = chunked_tokens_list
        batch_dict['chunked_tokens_lens'] = chunked_tokens_lens_list

        return batch_dict



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import json

import einops
import lhotse
import lightning.pytorch as pl
import numpy as np
import pytest
import soundfile as sf
import torch
from omegaconf import DictConfig

from nemo.collections.audio.models.enhancement import SchroedingerBridgeAudioToAudioModel


@pytest.fixture(params=["nemo_manifest", "lhotse_cuts"])
def mock_dataset_config(tmp_path, request):
    num_files = 8
    num_samples = 16000

    for i in range(num_files):
        data = np.random.randn(num_samples, 1)
        sf.write(tmp_path / f"audio_{i}.wav", data, 16000)

    if request.param == "lhotse_cuts":
        with lhotse.CutSet.open_writer(tmp_path / "cuts.jsonl") as writer:
            for i in range(num_files):
                recording = lhotse.Recording.from_file(tmp_path / f"audio_{i}.wav")
                cut = lhotse.MonoCut(
                    id=f"audio_{i}",
                    start=0,
                    channel=0,
                    duration=num_samples / 16000,
                    recording=recording,
                    custom={"target_recording": recording},
                )
                writer.write(cut)

            return {
                'cuts_path': str(tmp_path / "cuts.jsonl"),
                'use_lhotse': True,
                'batch_size': 2,
                'num_workers': 1,
            }
    elif request.param == "nemo_manifest":
        with (tmp_path / "small_manifest.jsonl").open("w") as f:
            for i in range(num_files):
                entry = {
                    "noisy_filepath": str(tmp_path / f"audio_{i}.wav"),
                    "clean_filepath": str(tmp_path / f"audio_{i}.wav"),
                    "duration": num_samples / 16000,
                    "offset": 0,
                }
                f.write(f"{json.dumps(entry)}\n")
        return {
            'manifest_filepath': str(tmp_path / "small_manifest.jsonl"),
            'input_key': 'noisy_filepath',
            'target_key': 'clean_filepath',
            'use_lhotse': False,
            'batch_size': 2,
            'num_workers': 1,
        }
    else:
        raise NotImplementedError(f"Dataset type {request.param} not implemented")


@pytest.fixture()
def schroedinger_bridge_model_ncsn_params():

    model = {
        'sample_rate': 16000,
        'num_outputs': 1,
        'normalize_input': True,
        'max_utts_evaluation_metrics': 50,
    }
    encoder = {
        '_target_': 'nemo.collections.audio.modules.transforms.AudioToSpectrogram',
        'fft_length': 510,
        'hop_length': 128,
        'magnitude_power': 0.5,
        'scale': 0.33,
    }
    decoder = {
        '_target_': 'nemo.collections.audio.modules.transforms.SpectrogramToAudio',
        'fft_length': encoder['fft_length'],
        'hop_length': encoder['hop_length'],
        'magnitude_power': encoder['magnitude_power'],
        'scale': encoder['scale'],
    }
    estimator = {
        '_target_': 'nemo.collections.audio.parts.submodules.ncsnpp.SpectrogramNoiseConditionalScoreNetworkPlusPlus',
        'in_channels': 2,  # single-channel noisy input
        'out_channels': 1,  # single-channel estimate
        'conditioned_on_time': True,
        'channels': [8, 8, 8, 8, 8],
        'num_res_blocks': 3,  # increased number of res blocks
        'pad_time_to': 64,  # pad to 64 frames for the time dimension
        'pad_dimension_to': 0,  # no padding in the frequency dimension
    }

    loss_encoded = {
        '_target_': 'nemo.collections.audio.losses.audio.MSELoss',
        'ndim': 4,
    }  # computed in the time domain

    loss_time = {'_target_': 'nemo.collections.audio.losses.audio.MAELoss'}

    noise_schedule = {
        '_target_': 'nemo.collections.audio.parts.submodules.schroedinger_bridge.SBNoiseScheduleVE',
        'k': 2.6,
        'c': 0.4,
        'time_min': 1e-4,
        'time_max': 1.0,
        'num_steps': 1000,  # num steps for the forward process
    }

    sampler = {
        '_target_': 'nemo.collections.audio.parts.submodules.schroedinger_bridge.SBSampler',
        'time_min': 1e-4,
        'time_max': 1.0,
        'num_steps': 5,  # num steps for the reverse process
    }

    metrics = {
        'val': {
            'sisdr': {
                '_target_': 'torchmetrics.audio.ScaleInvariantSignalDistortionRatio',
            },
        },
    }

    model_config = DictConfig(
        {
            'sample_rate': model['sample_rate'],
            'num_outputs': model['num_outputs'],
            'normalize_input': model['normalize_input'],
            'max_utts_evaluation_metrics': model['max_utts_evaluation_metrics'],
            'encoder': DictConfig(encoder),
            'decoder': DictConfig(decoder),
            'estimator': DictConfig(estimator),
            'loss_encoded': DictConfig(loss_encoded),
            'loss_time': DictConfig(loss_time),
            'loss_time_weight': 0.001,
            'estimator_output': 'data_prediction',
            'noise_schedule': DictConfig(noise_schedule),
            'sampler': DictConfig(sampler),
            'metrics': metrics,
            'optim': {
                'name': 'adam',
                'lr': 0.001,
                'betas': (0.9, 0.98),
            },
        }
    )

    return model_config


@pytest.fixture()
def schroedinger_bridge_model_ncsn(schroedinger_bridge_model_ncsn_params):
    with torch.random.fork_rng():
        torch.random.manual_seed(0)
        model = SchroedingerBridgeAudioToAudioModel(cfg=schroedinger_bridge_model_ncsn_params)
    return model


@pytest.fixture()
def schroedinger_bridge_model_ncsn_with_trainer_and_mock_dataset(
    schroedinger_bridge_model_ncsn_params, mock_dataset_config
):
    # Add train and validation dataset configs
    schroedinger_bridge_model_ncsn_params["train_ds"] = {**mock_dataset_config, "shuffle": True}
    schroedinger_bridge_model_ncsn_params["validation_ds"] = {**mock_dataset_config, "shuffle": False}

    # Trainer config
    trainer_cfg = {
        "max_epochs": -1,
        "max_steps": 8,
        "logger": False,
        "use_distributed_sampler": False,
        "val_check_interval": 2,
        "limit_train_batches": 4,
        "accelerator": "cpu",
        "enable_checkpointing": False,
    }
    schroedinger_bridge_model_ncsn_params["trainer"] = trainer_cfg

    trainer = pl.Trainer(**trainer_cfg)

    with torch.random.fork_rng():
        torch.random.manual_seed(0)
        model = SchroedingerBridgeAudioToAudioModel(cfg=schroedinger_bridge_model_ncsn_params, trainer=trainer)

    return model, trainer


class TestSchroedingerBridgeModelNCSN:
    """Test Schroedinger Bridge model with NCSN estimator."""

    @pytest.mark.unit
    def test_constructor(self, schroedinger_bridge_model_ncsn):
        """Test that the model can be constructed from a config dict."""
        model = schroedinger_bridge_model_ncsn.train()
        confdict = model.to_config_dict()
        instance2 = SchroedingerBridgeAudioToAudioModel.from_config_dict(confdict)
        assert isinstance(instance2, SchroedingerBridgeAudioToAudioModel)

    @pytest.mark.unit
    @pytest.mark.parametrize(
        "batch_size, sample_len",
        [
            (4, 4),  # Example 1
            (2, 8),  # Example 2
            (1, 10),  # Example 3
        ],
    )
    def test_forward_infer(self, schroedinger_bridge_model_ncsn, batch_size, sample_len):
        """Test that the model can run forward inference."""
        model = schroedinger_bridge_model_ncsn.eval()
        confdict = model.to_config_dict()
        sampling_rate = confdict['sample_rate']
        rng = torch.Generator()
        rng.manual_seed(0)
        input_signal = torch.randn(size=(batch_size, 1, sample_len * sampling_rate), generator=rng)
        input_signal_length = (sample_len * sampling_rate) * torch.ones(batch_size, dtype=torch.int)

        with torch.no_grad():
            # batch size 1
            output_list = []
            output_length_list = []
            for i in range(input_signal.size(0)):
                output, output_length = model.forward(
                    input_signal=input_signal[i : i + 1], input_length=input_signal_length[i : i + 1]
                )
                output_list.append(output)
                output_length_list.append(output_length)
            output_instance = torch.cat(output_list, 0)
            output_length_instance = torch.cat(output_length_list, 0)

            # batch size batch_size
            output_batch, output_length_batch = model.forward(
                input_signal=input_signal, input_length=input_signal_length
            )

        # It is generative model so we do not check the diffenence between output_instance and output_batch

        # Check that the output and output length are the same for the instance and batch
        assert output_instance.shape == output_batch.shape
        assert output_length_instance.shape == output_length_batch.shape

    def test_training_step(self, schroedinger_bridge_model_ncsn_with_trainer_and_mock_dataset):
        model, _ = schroedinger_bridge_model_ncsn_with_trainer_and_mock_dataset
        model = model.train()

        for batch in itertools.islice(model._train_dl, 2):
            # start boilerplate from SchroedingerBridgeAudioToAudioModel.training_step
            if isinstance(batch, dict):
                # lhotse batches are dictionaries
                input_signal = batch['input_signal']
                input_length = batch['input_length']
                target_signal = batch.get('target_signal', input_signal)
            else:
                input_signal, input_length, target_signal, _ = batch
            if input_signal.ndim == 2:
                input_signal = einops.rearrange(input_signal, 'B T -> B 1 T')
            if target_signal.ndim == 2:
                target_signal = einops.rearrange(target_signal, 'B T -> B 1 T')
            # end boilerplate

            loss, _, _ = model._step(target_signal=target_signal, input_signal=input_signal, input_length=input_length)
            loss.backward()

    def test_model_training(self, schroedinger_bridge_model_ncsn_with_trainer_and_mock_dataset):
        """
        Test that the model can be trained for a few steps. An evaluation step is also expected.
        """
        model, trainer = schroedinger_bridge_model_ncsn_with_trainer_and_mock_dataset
        model = model.train()

        trainer.fit(model)
